# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The source file of the class for the special quantum operator.
"""

from typing import Iterable, List, Optional, Union

import numpy as np
import torch

from ..core import Operator, OperatorInfoType, State, utils
from ..core.intrinsic import _alias, _digit_to_int, _int_to_digit
from . import Channel, Gate


class ResetState(Operator):
    r"""The class to reset the quantum state. It will be implemented soon.
    """
    def __init__(self):
        super().__init__()

    def forward(self, *inputs, **kwargs):
        r"""The forward function.

        Returns:
            NotImplemented.
        """
        return NotImplemented


class Collapse(Operator):
    r"""The class to compute the collapse of the quantum state.

    Args:
        system_idx: list of systems to be collapsed.
        acted_system_dim: dimension of systems that this gate acts on. Can be a list of system dimensions 
            or an int representing the dimension of all systems. Defaults to be qubit case. Used for collapse function.
        desired_result: The desired result you want to collapse. Defaults to ``None`` meaning preserving all results, 
            and activate probabilistic computation.
        if_print: whether print the information about the collapsed state. Defaults to ``False``.
        measure_basis: The basis of the measurement. Defaults to the computational basis.

    Raises:
        NotImplementedError: If the basis of measurement is not z. Other bases will be implemented in future.
    
    """
    @_alias({"system_idx": "qubits_idx"})
    def __init__(self, system_idx: Union[int, Iterable[int]], acted_system_dim: List[int], 
                 desired_result: Union[int, str] = None, if_print: bool = False, measure_basis: Optional[torch.Tensor] = None):
        super().__init__()
        self.measure_basis = []

        if isinstance(system_idx, int):
            self.system_idx = [system_idx]
        else:
            self.system_idx = list(system_idx)
        
        self.if_print = if_print
    
        if measure_basis:
            assert len(measure_basis) > 1, \
                "The input measurement op should be a list of measurement operators."
            assert utils.check.is_pvm(measure_basis), \
                "The input measurement op do not form a projection-valued measurement."
        self.measure_basis = measure_basis
        self.system_dim = acted_system_dim
        
        if desired_result is None:
            digits_str = None
        else:
            if isinstance(desired_result, str):
                desired_result = _digit_to_int(desired_result, acted_system_dim)
            digits_str = _int_to_digit(desired_result, acted_system_dim)
            desired_result = torch.tensor(desired_result)
        
        self.desired_result = desired_result
        self.digits_str = digits_str

        _info = {"name": "measure",
                 "system_idx": self.system_idx,
                 "type": "channel",
                 "kwargs": {"basis": measure_basis, "if_print": if_print}}
        if desired_result is not None:
            _info["label"] = digits_str
        self._info.update(_info)
        
    def __call__(self, state: State) -> State:
        r"""Same as forward of Neural Network
        """
        return self.forward(state)
    
    @property
    def info(self) -> OperatorInfoType:
        r"""Information of this channel
        """
        info = super().info
        info.update({
            'system_idx': self.system_idx,
        })
        return info
        
    def forward(self, state: State) -> State:
        r"""Compute the collapse of the input state.

        Args:
            state: The input state, which will be collapsed

        Returns:
            The collapsed quantum state.
        """
        prob_array, measured_state = state.measure(self.measure_basis, self.system_idx, keep_state=True)
        
        desired_result = self.desired_result
        if desired_result is None:
            return measured_state
        
        digits_str = self.digits_str
        state_str = '>'.join(f'|{d}' for d in digits_str) + '>'
        prob_collapse = prob_array.index_select(-1, desired_result)
        assert torch.all(prob_collapse > 1e-10).item(), (
            f"It is computationally infeasible for some states in systems {self.system_idx} "
            f"to collapse to state {state_str}")

        # whether print the collapsed result
        if self.if_print:
            prob = prob_collapse.mean().item()
            print(f"systems {self.system_idx} collapse to the state {state_str} with (average) probability {prob}")

        return measured_state.prob_select(desired_result)


class OneWayLOCC(Operator):
    r"""A one-way LOCC protocol, where quantum measurement is modelled by a PVM and all channels are unitary channels.
    
    Args:
        gate: a Gate operator.
        measure_basis: basis of the measurement. Defaults to the computational basis.
        label: name of the measured label. Defaults to ``'M'``.
        latex_name: latex name of the applied operator. Defaults to ``'O'``.
    
    """
    @_alias({'system_idx': 'qubits_idx'})
    def __init__(
        self, gate: Gate,
        measure_idx: List[int], measure_dim: Union[List[int], int],
        measure_basis: Optional[torch.Tensor] = None, label: str = 'M', latex_name: str = 'O'
    ):
        super().__init__()
        num_measure_system = len(measure_idx)
        measure_dim = [measure_dim] * num_measure_system if isinstance(measure_dim, int) else measure_dim
        
        assert (batch_dim := gate.matrix.shape[0]) == np.prod(measure_dim), \
            f"Batch dimension mismatch: expected {np.prod(measure_dim)}, received {batch_dim} unitaries"
        
        self.measure = Collapse(measure_idx, measure_dim, measure_basis=measure_basis)
        self.list_gate = gate
        
        self._info.update({'name': "locc",
                           'system_idx': measure_idx + gate.system_idx,
                           'tex': latex_name,
                           'label': label,
                           'type': "locc",
                           'num_ctrl_system': len(measure_idx),
                           'kwargs': {'measure_basis': measure_basis,
                                      'label_name': label}})
        
    def __call__(self, state: State) -> State:
        r"""Same as forward of Neural Network
        """
        return self.forward(state)
    
    @property
    def system_idx(self) -> List[int]:
        r"""The system indices of the LOCC protocol.
        """
        return self.measure.system_idx + self.list_gate.system_idx[0]
    
    @system_idx.setter
    def system_idx(self, value: List[int]):
        r"""Set the system indices of the LOCC protocol.
        """
        assert len(value) == len(self.system_idx), \
             f"Length of system_idx should be {len(self.system_idx)}, but got {len(value)}"
        num_measure_system = len(self.measure.system_idx)
        self.measure.system_idx = value[:num_measure_system]
        self.list_gate.system_idx = [value[num_measure_system:]]
    
    @property
    def info(self) -> OperatorInfoType:
        r"""Information of this channel
        """
        info = super().info
        info.update({
            'system_idx': self.system_idx,
        })
        return info
    
    def forward(self, state: State) -> State:
        r"""Compute the input state passing through the LOCC protocol.

        Args:
            state: The input state.

        Returns:
            The collapsed quantum state.
        
        """
        matrix, sys_idx = self.list_gate.matrix, self.list_gate.system_idx[0]
        
        measured_state = self.measure(state)
        
        dim = len(matrix.shape[:-2])
        matrix = matrix.repeat(measured_state._prob_dim[:-dim] + [1] * (dim + 2))
        measured_state._evolve(matrix, sys_idx, on_batch=False)
        return measured_state


class QuasiOperation(Operator):
    r"""A quantum protocol containing quasi-operations.
    
    Args:
        list_channels: a batched tensor that represents all unitaries.
        quasi_prob: the quasi-probability distribution for this quasi-operation.
        system_idx: indices of the systems on which the protocol is applied. 
        type_repr: one of ``'choi'``, ``'kraus'``, ``'stinespring'`` or ``'gate'``. Defaults to ``'gate'``.
        acted_system_dim: dimension of systems that these channels act on. Can be a list of system dimensions 
            or an int representing the dimension of all systems. Defaults to be qubit case.
    """
    def __init__(
        self, list_channels: torch.Tensor, quasi_prob: torch.Tensor, 
        system_idx: Union[Iterable[int], int] = None,
        type_repr: str = 'gate', acted_system_dim: Union[List[int], int] = 2
    ):
        super().__init__()
        
        # TODO: support other types of representations
        if type_repr != 'gate':
            raise NotImplementedError("Only 'gate' type is supported for now.")
        
        assert np.abs((s := quasi_prob.sum().item()) - 1) < 1e-4, \
            f"The quasi-probability distribution should sum to 1: received sum {s}"
        
        if type_repr == 'gate':
            self.channel = Gate(list_channels, system_idx, acted_system_dim)
        else:
            self.channel = Channel(type_repr, list_channels, system_idx, acted_system_dim)
        self.prob = quasi_prob.view([-1])
    
    def forward(self, state: State) -> State:
        r"""Compute the input state passing through the quasi-operation.
        
        Args:
            state: The input state.
        
        Returns:
            The collapsed quantum state.
        """
        state, prob = state.clone(), self.prob
        
        if state._prob:
            last_prob = state._prob[-1]
            prob = prob.view([1] * last_prob.ndim() + [-1])
        state._prob.append(self.prob)
        state._evolve(self.channel.matrix, self.channel.system_idx[0], on_batch=False)
        return state



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The library of common quantum states.
"""

import math
from typing import List, Union

import torch

from ..core import State, to_state
from ..core.intrinsic import _alias, _format_total_dim

__all__ = [
    "zero_state",
    "one_state",
    "computational_state",
    "bell_state",
    "bell_diagonal_state",
    "w_state",
    "ghz_state",
    "completely_mixed_computational",
    "r_state",
    "s_state",
    "isotropic_state",
]


@_alias({"num_systems": "num_qubits"})
def zero_state(num_systems: int, system_dim: Union[List[int], int] = 2) -> State:
    r"""Generate a zero state.

    Args:
        num_systems: Number of systems in this state. Alias of ``num_qubits``.
        system_dim: Dimension of systems. Can be a list of system dimensions 
            or an int representing the dimension of all systems. Defaults to qubit case.

    Returns:
        The generated quantum state.

    Examples:
        .. code-block:: python

            num_systems = 2
            system_dim = [2, 3]
            state = zero_state(num_systems, system_dim)
            print(f'The zero state is:\n{state}')

        ::

            The zero state is:
            ---------------------------------------------------
            Backend: state_vector
            System dimension: [2, 3]
            System sequence: [0, 1]
            [1.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j 0.+0.j]
            ---------------------------------------------------
    """
    return computational_state(num_systems, 0, system_dim)


@_alias({"num_systems": "num_qubits"})
def one_state(num_systems: int, system_dim: Union[List[int], int] = 2) -> State:
    r"""Generate a one state.

    Args:
        num_systems: Number of systems in this state. Alias of ``num_qubits``.
        system_dim: Dimension of systems. Can be a list of system dimensions 
            or an int representing the dimension of all systems. Defaults to qubit case.

    Returns:
        The generated quantum state.

    Examples:
        .. code-block:: python

            num_systems = 2
            system_dim = [1, 3]
            state = one_state(num_systems, system_dim)
            print(f'The one state is:\n{state}')

        ::

            The one state is:
            ---------------------------------------------------
            Backend: state_vector
            System dimension: [1, 3]
            System sequence: [0, 1]
            [0.+0.j 1.+0.j 0.+0.j]
            ---------------------------------------------------
    """
    return computational_state(num_systems, 1, system_dim)


@_alias({"num_systems": "num_qubits"})
def computational_state(num_systems: int, index: int, 
                        system_dim: Union[List[int], int] = 2) -> State:
    r"""Generate a computational state :math:`|e_{i}\rangle`, 
    whose i-th element is 1 and all the other elements are 0.

    Args:
        num_systems: Number of systems in this state. Alias of ``num_qubits``.
        index:  Index :math:`i` of the computational basis state :math:`|e_{i}\rangle`.
        system_dim: Dimension of systems. Can be a list of system dimensions 
            or an int representing the dimension of all systems. Defaults to qubit case.

    Returns:
        The generated quantum state.

    Examples:
        .. code-block:: python

            num_systems = 2
            system_dim = [2, 3]
            index = 4
            state = computational_state(num_systems, index, system_dim)
            print(f'The state is:\n{state}')

        ::

            The state is:
            ---------------------------------------------------
            Backend: state_vector
            System dimension: [2, 3]
            System sequence: [0, 1]
            [0.+0.j 0.+0.j 0.+0.j 0.+0.j 1.+0.j 0.+0.j]
            ---------------------------------------------------
    """
    dim = _format_total_dim(num_systems, system_dim)
    
    data = torch.zeros(dim)
    data[index] = 1
    return to_state(data, system_dim)


@_alias({"num_systems": "num_qubits"})
def bell_state(num_systems: int, system_dim: Union[List[int], int] = 2) -> State:
    r"""Generate a Bell state.

    Its matrix form is:

    .. math::

        |\Phi_{D}\rangle=\frac{1}{\sqrt{D}} \sum_{j=0}^{D-1}|j\rangle_{A}|j\rangle_{B}

    Args:
        num_systems: Number of systems in this state. Alias of ``num_qubits``.
        system_dim: Dimension of systems. Can be a list of system dimensions 
            or an int representing the dimension of all systems. Defaults to qubit case.

    Returns:
        The generated quantum state.

    Examples:
        .. code-block:: python

            num_systems = 2
            system_dim = [2, 2]
            state = bell_state(num_systems, system_dim)
            print(f'The Bell state is:\n{state}')

        ::

            The Bell state is:
            ---------------------------------------------------
            Backend: state_vector
            System dimension: [2, 2]
            System sequence: [0, 1]
            [0.71+0.j 0.  +0.j 0.  +0.j 0.71+0.j]
            ---------------------------------------------------
    """
    assert num_systems % 2 == 0, \
        f"Number of systems must be even to form a Bell state. Received: {num_systems}"
    half = num_systems // 2
    
    dim = _format_total_dim(num_systems, system_dim)
    if isinstance(system_dim, int):
        local_dim = system_dim ** half
    else:
        local_dim = math.prod(system_dim[:half])
        assert dim == (local_dim ** 2), \
            f"Dimension of systems must be evenly distributed. Received: {system_dim}"
    
    data = torch.zeros(dim)
    for i in range(0, dim, local_dim + 1 ):
        data[i] = 1 / math.sqrt(local_dim)
    return to_state(data, system_dim)


def bell_diagonal_state(prob: List[float]) -> State:
    r"""Generate a Bell diagonal state.

    Its matrix form is:

    .. math::

        p_{1}|\Phi^{+}\rangle\langle\Phi^{+}|+p_{2}|\Psi^{+}\rangle\langle\Psi^{+}|+p_{3}|\Phi^{-}\rangle\langle\Phi^{-}| +
        p_{4}|\Psi^{-}\rangle\langle\Psi^{-}|

    Args:
        prob: The probability of each Bell state.

    Raises:
        Exception: The state should be a pure state if the backend is state_vector.
        NotImplementedError: If the backend is wrong or not implemented.

    Returns:
        The generated quantum state.

    Examples:
        .. code-block:: python

            prob = [0.2, 0.3, 0.4, 0.1]
            state = bell_diagonal_state(prob)
            print(f'The Bell diagonal state is:\n{state}')

        ::

            The Bell diagonal state is:
            ---------------------------------------------------
            Backend: density_matrix
            System dimension: [2, 2]
            System sequence: [0, 1]
            [[ 0.3+0.j  0. +0.j  0. +0.j -0.1+0.j]
             [ 0. +0.j  0.2+0.j  0.1+0.j  0. +0.j]
             [ 0. +0.j  0.1+0.j  0.2+0.j  0. +0.j]
             [-0.1+0.j  0. +0.j  0. +0.j  0.3+0.j]]
            ---------------------------------------------------
    """
    p1, p2, p3, p4 = prob
    assert 0 <= p1 <= 1 and 0 <= p2 <= 1 and 0 <= p3 <= 1 and 0 <= p4 <= 1, \
        "Each probability must be in [0, 1]."
    assert abs(p1 + p2 + p3 + p4 - 1) < 1e-6, \
        "The sum of probabilities should be 1."

    phi_plus = torch.tensor([[1, 0, 0, 1],
                             [0, 0, 0, 0],
                             [0, 0, 0, 0],
                             [1, 0, 0, 1]]) / 2
    phi_minus = torch.tensor([[1, 0, 0, -1],
                              [0, 0, 0, 0],
                              [0, 0, 0, 0],
                              [-1, 0, 0, 1]]) / 2
    
    psi_plus = torch.tensor([[0, 0, 0, 0],
                             [0, 1, 1, 0],
                             [0, 1, 1, 0],
                             [0, 0, 0, 0]]) / 2
    psi_minus = torch.tensor([[0, 0, 0, 0],
                              [0, 1, -1, 0],
                              [0, -1, 1, 0],
                              [0, 0, 0, 0]]) / 2

    data = p1 * phi_plus + p2 * psi_plus + p3 * phi_minus + p4 * psi_minus

    return to_state(data)


def w_state(num_qubits: int) -> State:
    r"""Generate a W-state.

    Args:
        num_qubits: The number of qubits in the quantum state.

    Raises:
        NotImplementedError: If the backend is wrong or not implemented.

    Returns:
        The generated quantum state.

    Examples:
        .. code-block:: python

            num_qubits = 2
            w_state_inst = w_state(num_qubits)
            print(f'The W-state is:\n{w_state_inst}')

        ::

            The W-state is:
            ---------------------------------------------------
            Backend: state_vector
            System dimension: [2, 2]
            System sequence: [0, 1]
            [0.  +0.j 0.71+0.j 0.71+0.j 0.  +0.j]
            ---------------------------------------------------
    """
    dim = 2 ** num_qubits
    coeff = 1 / math.sqrt(num_qubits)
    data = torch.zeros(dim)

    for i in range(num_qubits):
        data[2 ** i] = coeff

    return to_state(data)


def ghz_state(num_qubits: int) -> State:
    r"""Generate a GHZ-state.

    Args:
        num_qubits: The number of qubits in the quantum state.

    Raises:
        NotImplementedError: If the backend is wrong or not implemented.

    Returns:
        The generated quantum state.

    Examples:
        .. code-block:: python

            num_qubits = 2
            ghz_state_inst = ghz_state(num_qubits)
            print(f'The GHZ-state is:\n{ghz_state_inst}')

        ::

            The GHZ-state is:
            ---------------------------------------------------
            Backend: state_vector
            System dimension: [2, 2]
            System sequence: [0, 1]
            [0.71+0.j 0.  +0.j 0.  +0.j 0.71+0.j]
            ---------------------------------------------------
    """
    dim = 2 ** num_qubits
    data = torch.zeros(dim)
    data[0] = 1 / math.sqrt(2)
    data[-1] = 1 / math.sqrt(2)
    return to_state(data)


def completely_mixed_computational(num_qubits: int) -> State:
    r"""Generate the density matrix of the completely mixed state.

    Args:
        num_qubits: The number of qubits in the quantum state.

    Raises:
        Exception: The state should be a pure state if the backend is state_vector.
        NotImplementedError: If the backend is wrong or not implemented.

    Returns:
        The generated quantum state.

    Examples:
        .. code-block:: python

            num_qubits = 1
            state = completely_mixed_computational(num_qubits)
            print(f'The density matrix of the completely mixed state is:\n{state}')

        ::

            The density matrix of the completely mixed state is:
            ---------------------------------------------------
            Backend: density_matrix
            System dimension: [2]
            System sequence: [0]
            [[0.5+0.j 0. +0.j]
             [0. +0.j 0.5+0.j]]
            ---------------------------------------------------
    """
    data = torch.eye(2 ** num_qubits) / (2 ** num_qubits)
    return to_state(data)


def r_state(prob: float) -> State:
    r"""Generate an R-state.

    Its matrix form is:

    .. math::

        p|\Psi^{+}\rangle\langle\Psi^{+}| + (1 - p)|11\rangle\langle11|

    Args:
        prob: The parameter of the R-state to be generated. It should be in :math:`[0,1]`.

    Raises:
        Exception: The state should be a pure state if the backend is state_vector.
        NotImplementedError: If the backend is wrong or not implemented.

    Returns:
        The generated quantum state.

    Examples:
        .. code-block:: python

            prob = 0.5
            r_state_inst = r_state(prob)
            print(f'The R-state is:\n{r_state_inst}')

        ::

            The R-state is:
            ---------------------------------------------------
            Backend: density_matrix
            System dimension: [2, 2]
            System sequence: [0, 1]
            [[0.  +0.j 0.  +0.j 0.  +0.j 0.  +0.j]
             [0.  +0.j 0.25+0.j 0.25+0.j 0.  +0.j]
             [0.  +0.j 0.25+0.j 0.25+0.j 0.  +0.j]
             [0.  +0.j 0.  +0.j 0.  +0.j 0.5 +0.j]]
            ---------------------------------------------------
    """
    assert 0 <= prob <= 1, "Probability must be in [0, 1]"

    phi_plus = torch.tensor([[0, 0, 0, 0],
                             [0, 1, 1, 0],
                             [0, 1, 1, 0],
                             [0, 0, 0, 0]]) / 2
    
    state_11 = torch.zeros((4, 4))
    state_11[3, 3] = 1
    data = prob * phi_plus + (1 - prob) * state_11
    
    return to_state(data)


def s_state(prob: float) -> State:
    r"""Generate the S-state.

    Its matrix form is:

    .. math::

        p|\Phi^{+}\rangle\langle\Phi^{+}| + (1 - p)|00\rangle\langle00|

    Args:
        prob: The parameter of the S-state to be generated. It should be in :math:`[0,1]`.

    Raises:
        Exception: The state should be a pure state if the backend is state_vector.
        NotImplementedError: If the backend is wrong or not implemented.

    Returns:
        The generated quantum state.

    Examples:
        .. code-block:: python

            prob = 0.5
            s_state_inst = s_state(prob)
            print(f'The S-state is:\n{s_state_inst}')

        ::

            The S-state is:
            ---------------------------------------------------
            Backend: density_matrix
            System dimension: [2, 2]
            System sequence: [0, 1]
            [[0.75+0.j 0.  +0.j 0.  +0.j 0.25+0.j]
             [0.  +0.j 0.  +0.j 0.  +0.j 0.  +0.j]
             [0.  +0.j 0.  +0.j 0.  +0.j 0.  +0.j]
             [0.25+0.j 0.  +0.j 0.  +0.j 0.25+0.j]]
            ---------------------------------------------------
    """
    assert 0 <= prob <= 1, "Probability must be in [0, 1]"
    
    phi_p = bell_state(2).density_matrix
    psi0 = torch.zeros_like(phi_p)
    psi0[0, 0] = 1
    data = prob * phi_p + (1 - prob) * psi0

    return to_state(data)


def isotropic_state(num_qubits: int, prob: float) -> State:
    r"""Generate the isotropic state.

    Its matrix form is:

    .. math::

        p\left(\frac{1}{\sqrt{D}} \sum_{j=0}^{D-1}|j\rangle_{A}|j\rangle_{B}\right) + (1 - p)\frac{I}{2^n}

    Args:
        num_qubits: The number of qubits in the quantum state.
        prob: The parameter of the isotropic state to be generated. It should be in :math:`[0,1]`.

    Raises:
        Exception: The state should be a pure state if the backend is state_vector.
        NotImplementedError: If the backend is wrong or not implemented.

    Returns:
        The generated quantum state.

    Examples:
        .. code-block:: python

            num_qubits = 2
            prob = 0.5
            isotropic_state_inst = isotropic_state(num_qubits, prob)
            print(f'The isotropic state is:\n{isotropic_state_inst}')

        ::

            The isotropic state is:
            ---------------------------------------------------
            Backend: density_matrix
            System dimension: [2, 2]
            System sequence: [0, 1]
            [[0.38+0.j 0.  +0.j 0.  +0.j 0.25+0.j]
             [0.  +0.j 0.12+0.j 0.  +0.j 0.  +0.j]
             [0.  +0.j 0.  +0.j 0.12+0.j 0.  +0.j]
             [0.25+0.j 0.  +0.j 0.  +0.j 0.38+0.j]]
            ---------------------------------------------------
    """
    assert 0 <= prob <= 1, "Probability must be in [0, 1]"

    dim = 2 ** num_qubits
    phi_b = bell_state(num_qubits).density_matrix
    data = prob * phi_b + (1 - prob) * torch.eye(dim) / dim

    return to_state(data)



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The source file of the Circuit class.
"""

import math
import shutil
import warnings
from typing import Callable, Dict, Iterable, List, Optional, Tuple, Union

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import torch

from .ansatz import (ComplexBlockLayer, ComplexEntangledLayer, Layer,
                     LinearEntangledLayer, OperatorList, RealBlockLayer,
                     RealEntangledLayer, TrotterLayer, Universal2, Universal3)
from .core import Hamiltonian, intrinsic, latex, utils
from .core.intrinsic import _alias
from .operator import (CCX, CNOT, CP, CRX, CRY, CRZ, CSWAP, CU, CY, CZ, MS, RX,
                       RXX, RY, RYY, RZ, RZZ, SWAP, U3, AmplitudeDamping,
                       BitFlip, BitPhaseFlip, ChoiRepr, Collapse,
                       ControlOracle, ControlParamOracle, Depolarizing, Gate,
                       GeneralizedAmplitudeDamping, GeneralizedDepolarizing, H,
                       KrausRepr, OneWayLOCC, Oracle, P, ParamOracle,
                       PauliChannel, Permutation, PhaseDamping, PhaseFlip,
                       ResetChannel, S, Sdg, StinespringRepr, T, Tdg,
                       ThermalRelaxation, UniversalQudits, X, Y, Z)
from .operator.gate import _circuit_plot

__all__ = ['Circuit']


class Circuit(OperatorList):
    r"""Class for quantum circuit.

    Args:
        num_systems: number of systems in the circuit. Defaults to None. Alias of ``num_qubits``.
        system_dim: dimension of systems of this circuit. Can be a list of system dimensions 
            or an int representing the dimension of all systems. Defaults to be qubit case.
        physical_idx: physical indices of systems. Defaults to be the same as the logical indices.
    
    Note:
        when the number of system is unknown and system_dim is an int, the circuit is a dynamic quantum circuit.

    Examples:
        .. code-block:: python

            qc = Circuit(1)  # A quantum circuit with 1 qubit
            qc.h()
            print(f'The latex code of this circuit is:\n{qc.to_latex()}')

        ::

            The latex code of this circuit is:
            \lstick{} & \gate[1]{H}
    """
    @_alias({'num_systems': 'num_qubits'})
    def __init__(self, num_systems: Optional[int] = None, 
                 system_dim: Optional[Union[List[int], int]] = 2,
                 physical_idx: Optional[List[int]] = None) -> None:
        super().__init__(num_systems, system_dim, physical_idx)

        # alias
        self.toffoli = self.ccx
        self.cx = self.cnot
        self.collapse = self.measure

    def _get_drawer(self, style: str, decimal: int) -> Tuple[latex.OperatorListDrawer, Dict[int, str]]:
        r"""Return the drawer that draws the circuit in LaTeX format, and the beginning string of the circuit.

        Examples:
            .. code-block:: python

                drawer, begin_code = qc._Circuit__get_drawer('standard', 2)

            ::

                drawer: (an instance of latex.OperatorListDrawer)
                begin_code: {0: r'\lstick{}', 1: r'\lstick{}', ...}
        """
        physical_idx, self.system_idx = self.system_idx, list(range(self.num_systems))
        
        drawer = latex.OperatorListDrawer(style, decimal)
        
        for op in self.children():
            if isinstance(op, Layer):
                drawer = drawer + op._get_drawer(style, decimal)
            else:
                drawer.append(op.info)
        
        begin_code = {idx: r'\lstick{}' for idx in self.system_idx}
        self.system_idx = physical_idx
        return drawer, begin_code
        
    def to_latex(self, style: str = 'standard', decimal: int = 2) -> str:
        r"""The LaTeX representation of the circuit, written in Quantikz format.
        
        Args:
            style: the style of the plot, can be 'standard', 'compact' or 'detailed'. Defaults to ``standard``.
            decimal: number of decimal places to display. Defaults to 2.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.h()
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[1]{H} & \meter[2]{} & {} \\
                \lstick{} & \gate[1]{H} & {} & {}
        """
        physical_idx, self.system_idx = self.system_idx, list(range(self.num_systems))
        drawer, begin_code = self._get_drawer(style, decimal)
        drawer.fill_all()
        self.system_idx = physical_idx
        return latex.code_to_str(drawer.code, begin_code)
    
    def plot(self, style: str = 'standard', decimal: int = 2,
             dpi: int = 300, print_code: bool = False, 
             show_plot: bool = True, include_empty: bool = False,
             latex: bool = True, **kwargs) -> Optional[matplotlib.figure.Figure]:
        r'''Display the circuit using Quantikz if ``latex`` is True, otherwise using matplotlib.

        Args:
            style: the style of the plot, can be 'standard', 'compact' or 'detailed'. Defaults to ``standard``.
            decimal: number of decimal places to display. Defaults to 2.
            dpi: dots per inches of plot image. Defaults to 300.
            print_code: whether print the LaTeX code of the circuit, default to ``False``.
            show_plot: whether show the plotted circuit, default to ``True``.
            include_empty: whether include empty lines, default to ``False``.
            latex: whether use Quantikz, a LaTeX package, to plot circuits , default to ``True``.
            kwargs: additional parameters for matplotlib plot.
        
        Returns:
            None, or a ``matplotlib.figure.Figure`` instance depending on ``latex`` and ``output``.

        Notes:
            If ``latex`` is True, the circuit will be displayed in LaTeX format;
            if ``latex`` is False, the circuit will be displayed in matplotlib format,
            in which case we have three additional parameters:
            - output_plot: whether output the plot instance, default to ``False``.
            - save_path: the save path of image. Defaults to None.
            - scale: scale coefficient of figure. Default to 1.0.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.h()
                qc.plot()

            ::

                (Displays the plotted circuit using Quantikz or matplotlib)
                
        .. warning::
        
            Starting from QuAIRKit 2.5.0, `Circuit.plot` now defaults to using QuantiKz, 
            a LaTeX package powered by TikZ, for rendering quantum circuit diagrams 
            commonly used in academic publications. Support for plotting circuits with 
            Matplotlib will be deprecated in the near future.

            To fully utilize this feature, please ensure that a TeX distribution, such 
            as [TeX Live](https://www.tug.org/texlive), is installed on your system. 
            This will enhance your experience with QuAIRKit and quantum computing 
            visualization.
        '''
        if latex and (shutil.which("pdflatex") is None):
            warnings.warn(
                "pdflatex is mot detected on your system. Will skip the plot.", UserWarning)
            
            if print_code:
                print(self.to_latex(style, decimal))
            return

        if latex:
            drawer, begin_code = self._get_drawer(style, decimal)   
            drawer._fill_empty(list(range(self.num_systems if include_empty else max(drawer._code.keys()))))
            drawer.add_end()
            _fig = drawer.plot(dpi, print_code, begin_code)
            
            if show_plot:
                intrinsic._display_png(_fig)
        else:
            warnings.warn(
                "Starting from QuAIRKit 0.4.0, `Circuit.plot` now defaults to using Quantikz, a LaTeX package "
                "powered by TikZ, for rendering quantum circuit diagrams commonly used in academic publications. "
                "Support for plotting circuits with Matplotlib will be deprecated in the near future. \n"
                "To fully utilize this feature, please ensure that a TeX distribution, such as TeX Live "
                "(https://www.tug.org/texlive), is installed on your system. This will enhance your experience "
                "with QuAIRKit and quantum computing visualization.", FutureWarning)
            save_path, scale = kwargs.get('save_path', None), kwargs.get('scale', 1.0)
            output_plot = kwargs.get('output_plot', False)
            
            physical_idx, self.system_idx = self.system_idx, list(range(self.num_systems))
            _fig = _circuit_plot(self, dpi=dpi, scale=scale)
            self.system_idx = physical_idx
            
            if save_path:
                plt.savefig(save_path, dpi=dpi)
            if show_plot:
                plt.show()
            if output_plot:
                return _fig


    # ---------------------- below are common operators ---------------------- #
    def h(self, qubits_idx: Union[Iterable[int], int, str] = 'full') -> None:
        r"""Add single-qubit Hadamard gates.

        The matrix form of such a gate is:

        .. math::
            H = \frac{1}{\sqrt{2}}
                \begin{bmatrix}
                    1 &  1 \\
                    1 & -1
                \end{bmatrix}

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied. Defaults to 'full'.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.h()
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[1]{H} & \meter[2]{} & {} \\
                \lstick{} & \gate[1]{H} & {} & {}
        """
        qubits_idx = self.register_idx(qubits_idx, 1)
        self.append(H(qubits_idx))

    def s(self, qubits_idx: Union[Iterable[int], int, str] = 'full') -> None:
        r"""Add single-qubit S gates.

        The matrix form of such a gate is:

        .. math::
            S =
                \begin{bmatrix}
                    1 & 0 \\
                    0 & i
                \end{bmatrix}

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied. Defaults to 'full'.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.s()
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[1]{S} & \meter[2]{} & {} \\
                \lstick{} & \gate[1]{S} & {} & {}
        """
        qubits_idx = self.register_idx(qubits_idx, 1)
        self.append(S(qubits_idx))

    def sdg(self, qubits_idx: Union[Iterable[int], int, str] = 'full') -> None:
        r"""Add single-qubit S dagger (S inverse) gates.

        The matrix form of such a gate is:

        .. math::
            S^\dagger =
                \begin{bmatrix}
                    1 &  0 \\
                    0 & -i
                \end{bmatrix}

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied. Defaults to ``'full'``.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.sdg()
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[1]{S^\dagger} & \meter[2]{} & {} \\
                \lstick{} & \gate[1]{S^\dagger} & {} & {}
        """
        qubits_idx = self.register_idx(qubits_idx, 1)
        self.append(Sdg(qubits_idx))

    def t(self, qubits_idx: Union[Iterable[int], int, str] = 'full') -> None:
        r"""Add single-qubit T gates.

        The matrix form of such a gate is:

        .. math::
            T = \begin{bmatrix}
                    1 & 0 \\
                    0 & e^{i\pi/4}
                \end{bmatrix}

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied. Defaults to 'full'.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.t()
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[1]{T} & \meter[2]{} & {} \\
                \lstick{} & \gate[1]{T} & {} & {}
        """
        qubits_idx = self.register_idx(qubits_idx, 1)
        self.append(T(qubits_idx))

    def tdg(self, qubits_idx: Union[Iterable[int], int, str] = 'full') -> None:
        r"""Add single-qubit T dagger (T inverse) gates.

        The matrix form of such a gate is:

        .. math::
            T^\dagger = \begin{bmatrix}
                            1 & 0 \\
                            0 & e^{-i\pi/4}
                         \end{bmatrix}

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied. Defaults to ``'full'``.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.tdg()
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[1]{T^\dagger} & \meter[2]{} & {} \\
                \lstick{} & \gate[1]{T^\dagger} & {} & {}
        """
        qubits_idx = self.register_idx(qubits_idx, 1)
        self.append(Tdg(qubits_idx))

    def x(self, qubits_idx: Union[Iterable[int], int, str] = 'full') -> None:
        r"""Add single-qubit X gates.

        The matrix form of such a gate is:

        .. math::
           X = \begin{bmatrix}
                    0 & 1 \\
                    1 & 0
               \end{bmatrix}

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied. Defaults to 'full'.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.x()
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[1]{X} & \meter[2]{} & {} \\
                \lstick{} & \gate[1]{X} & {} & {}
        """
        qubits_idx = self.register_idx(qubits_idx, 1)
        self.append(X(qubits_idx))

    def y(self, qubits_idx: Union[Iterable[int], int, str] = 'full') -> None:
        r"""Add single-qubit Y gates.

        The matrix form of such a gate is:

        .. math::
            Y = \begin{bmatrix}
                    0 & -i \\
                    i &  0
                \end{bmatrix}

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied. Defaults to 'full'.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.y()
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[1]{Y} & \meter[2]{} & {} \\
                \lstick{} & \gate[1]{Y} & {} & {}
        """
        qubits_idx = self.register_idx(qubits_idx, 1)
        self.append(Y(qubits_idx))

    def z(self, qubits_idx: Union[Iterable[int], int, str] = 'full') -> None:
        r"""Add single-qubit Z gates.

        The matrix form of such a gate is:

        .. math::
            Z = \begin{bmatrix}
                    1 &  0 \\
                    0 & -1
                \end{bmatrix}

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied. Defaults to 'full'.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.z()
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[1]{Z} & \meter[2]{} & {} \\
                \lstick{} & \gate[1]{Z} & {} & {}
        """
        qubits_idx = self.register_idx(qubits_idx, 1)
        self.append(Z(qubits_idx))

    def p(self, qubits_idx: Union[Iterable[int], int, str] = 'full',
          param: Union[torch.Tensor, float] = None, param_sharing: bool = False) -> None:
        r"""Add single-qubit P gates.

        The matrix form of such a gate is:

        .. math::
            P(\theta) = \begin{bmatrix}
                            1 & 0 \\
                            0 & e^{i\theta}
                        \end{bmatrix}

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied. Defaults to 'full'.
            param: Parameters of the gates. Defaults to None.
            param_sharing: Whether gates in the same layer share a parameter. Defaults to False.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.p([0], torch.pi/2)
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[1]{P(1.57)} & \meter[2]{} & {} \\
                \lstick{} & {} & {} & {}
        """
        qubits_idx = self.register_idx(qubits_idx, 1)
        self.append(P(qubits_idx, param, param_sharing))

    def rx(self, qubits_idx: Union[Iterable[int], int, str] = 'full',
           param: Union[torch.Tensor, float] = None, param_sharing: bool = False) -> None:
        r"""Add single-qubit rotation gates about the x-axis.

        The matrix form of such a gate is:

        .. math::
            R_X(\theta) = \begin{bmatrix}
                                \cos\frac{\theta}{2} & -i\sin\frac{\theta}{2} \\
                                -i\sin\frac{\theta}{2} & \cos\frac{\theta}{2}
                           \end{bmatrix}

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied. Defaults to 'full'.
            param: Parameters of the gates. Defaults to None.
            param_sharing: Whether gates in the same layer share a parameter. Defaults to False.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.rx([0], torch.pi/2)
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[1]{R_{x}(1.57)} & \meter[2]{} & {} \\
                \lstick{} & {} & {} & {}
        """
        qubits_idx = self.register_idx(qubits_idx, 1)
        self.append(RX(qubits_idx, param, param_sharing))

    def ry(self, qubits_idx: Union[Iterable[int], int, str] = 'full',
           param: Union[torch.Tensor, float] = None, param_sharing: bool = False) -> None:
        r"""Add single-qubit rotation gates about the y-axis.

        The matrix form of such a gate is:

        .. math::
            R_Y(\theta) = \begin{bmatrix}
                                \cos\frac{\theta}{2} & -\sin\frac{\theta}{2} \\
                                \sin\frac{\theta}{2} & \cos\frac{\theta}{2}
                           \end{bmatrix}

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied. Defaults to 'full'.
            param: Parameters of the gates. Defaults to None.
            param_sharing: Whether gates in the same layer share a parameter. Defaults to False.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.ry([0], torch.pi/2)
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[1]{R_{y}(1.57)} & \meter[2]{} & {} \\
                \lstick{} & {} & {} & {}
        """
        qubits_idx = self.register_idx(qubits_idx, 1)
        self.append(RY(qubits_idx, param, param_sharing))

    def rz(self, qubits_idx: Union[Iterable[int], int, str] = 'full',
           param: Union[torch.Tensor, float] = None, param_sharing: bool = False) -> None:
        r"""Add single-qubit rotation gates about the z-axis.

        The matrix form of such a gate is:

        .. math::
            R_Z(\theta) = \begin{bmatrix}
                                e^{-i\theta/2} & 0 \\
                                0 & e^{i\theta/2}
                           \end{bmatrix}

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied. Defaults to 'full'.
            param: Parameters of the gates. Defaults to None.
            param_sharing: Whether gates in the same layer share a parameter. Defaults to False.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.rz([0], torch.pi/2)
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[1]{R_{z}(1.57)} & \meter[2]{} & {} \\
                \lstick{} & {} & {} & {}
        """
        qubits_idx = self.register_idx(qubits_idx, 1)
        self.append(RZ(qubits_idx, param, param_sharing))

    def u3(self, qubits_idx: Union[Iterable[int], int, str] = 'full',
           param: Union[torch.Tensor, Iterable[float]] = None, param_sharing: bool = False) -> None:
        r"""Add single-qubit rotation gates.

        The matrix form of such a gate is:

        .. math::
            U_3(\theta, \phi, \lambda) =
                \begin{bmatrix}
                    \cos\frac\theta2 & -e^{i\lambda}\sin\frac\theta2 \\
                    e^{i\phi}\sin\frac\theta2 & e^{i(\phi+\lambda)}\cos\frac\theta2
                \end{bmatrix}

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied. Defaults to 'full'.
            param: Parameters of the gates. Defaults to None.
            param_sharing: Whether gates in the same layer share a parameter. Defaults to False.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.u3([0], torch.pi/2)
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[1]{U(1.57, 1.57, 1.57)} & \meter[2]{} & {} \\
                \lstick{} & {} & {} & {}
        """
        qubits_idx = self.register_idx(qubits_idx, 1)
        self.append(U3(qubits_idx, param, param_sharing))

    def cnot(self, qubits_idx: Union[Iterable[int], str] = 'cycle') -> None:
        r"""Add CNOT gates.

        For a 2-qubit circuit, when `qubits_idx` is `[0, 1]`, the matrix form is:

        .. math::
            \mathit{CNOT} = |0\rangle \langle 0|\otimes I + |1\rangle \langle 1|\otimes X

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied. Defaults to 'cycle'.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.cnot([0, 1])
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \ctrl[]{1} & \meter[2]{} & {} \\
                \lstick{} & \targ{} & {} & {}
        """
        qubits_idx = self.register_idx(qubits_idx, 2)
        self.append(CNOT(qubits_idx))

    def cy(self, qubits_idx: Union[Iterable[int], str] = 'cycle') -> None:
        r"""Add controlled Y gates.

        For a 2-qubit circuit, when `qubits_idx` is `[0, 1]`, the matrix form is:

        .. math::
            \mathit{CY} = |0\rangle \langle 0|\otimes I + |1\rangle \langle 1|\otimes Y

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied. Defaults to 'cycle'.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.cy([0, 1])
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \ctrl[]{1} & \meter[2]{} & {} \\
                \lstick{} & \gate[1]{Y} & {} & {}
        """
        qubits_idx = self.register_idx(qubits_idx, 2)
        self.append(CY(qubits_idx))

    def cz(self, qubits_idx: Union[Iterable[int], str] = 'linear') -> None:
        r"""Add controlled Z gates.

        For a 2-qubit circuit, when `qubits_idx` is `[0, 1]`, the matrix form is:

        .. math::
            \mathit{CZ} = |0\rangle \langle 0|\otimes I + |1\rangle \langle 1|\otimes Z

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied. Defaults to 'linear'.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.cz([0, 1])
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \ctrl[]{1} & \meter[2]{} & {} \\
                \lstick{} & \gate[1]{Z} & {} & {}
        """
        qubits_idx = self.register_idx(qubits_idx, 2)
        self.append(CZ(qubits_idx))

    def swap(self, qubits_idx: Union[Iterable[int], str] = 'linear') -> None:
        r"""Add SWAP gates.

        The matrix form is:

        .. math::
            \mathit{SWAP} =
            \begin{bmatrix}
                1 & 0 & 0 & 0 \\
                0 & 0 & 1 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & 0 & 1
            \end{bmatrix}

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied. Defaults to 'linear'.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.swap([0, 1])
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[2,style={draw=none}]{\permute{2,1}} & \meter[2]{} & {} \\
                \lstick{} & {} & {} & {}
        """
        qubits_idx = self.register_idx(qubits_idx, 2)
        self.append(SWAP(qubits_idx))

    def cp(self, qubits_idx: Union[Iterable[int], str] = 'cycle',
           param: Union[torch.Tensor, float] = None, param_sharing: bool = False) -> None:
        r"""Add controlled P gates.

        For a 2-qubit circuit, when `qubits_idx` is `[0, 1]`, the matrix form is:

        .. math::
            \mathit{CP}(\theta) =
            \begin{bmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & 1 & 0 \\
                0 & 0 & 0 & e^{i\theta}
            \end{bmatrix}

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied. Defaults to 'cycle'.
            param: Parameters of the gates. Defaults to None.
            param_sharing: Whether gates in the same layer share a parameter. Defaults to False.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.cp([0, 1], torch.pi / 2)
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \ctrl[]{1} & \meter[2]{} & {} \\
                \lstick{} & \gate[1]{P(1.57)} & {} & {}
        """
        qubits_idx = self.register_idx(qubits_idx, 2)
        self.append(CP(qubits_idx, param, param_sharing))

    def crx(self, qubits_idx: Union[Iterable[int], str] = 'cycle',
            param: Union[torch.Tensor, float] = None, param_sharing: bool = False) -> None:
        r"""Add controlled rotation gates about the x-axis.

        For a 2-qubit circuit, when `qubits_idx` is `[0, 1]`, the matrix form is:

        .. math::
            \mathit{CR_X} =
            |0\rangle \langle 0|\otimes I + |1\rangle \langle 1|\otimes R_X

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied. Defaults to 'cycle'.
            param: Parameters of the gates. Defaults to None.
            param_sharing: Whether gates in the same layer share a parameter. Defaults to False.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.crx([0, 1], torch.pi / 2)
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \ctrl[]{1} & \meter[2]{} & {} \\
                \lstick{} & \gate[1]{R_{x}(1.57)} & {} & {}
        """
        qubits_idx = self.register_idx(qubits_idx, 2)
        self.append(CRX(qubits_idx, param, param_sharing))

    def cry(self, qubits_idx: Union[Iterable[int], str] = 'cycle',
            param: Union[torch.Tensor, float] = None, param_sharing: bool = False) -> None:
        r"""Add controlled rotation gates about the y-axis.

        For a 2-qubit circuit, when `qubits_idx` is `[0, 1]`, the matrix form is:

        .. math::
            \mathit{CR_Y} =
            |0\rangle \langle 0|\otimes I + |1\rangle \langle 1|\otimes R_Y

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied. Defaults to 'cycle'.
            param: Parameters of the gates. Defaults to None.
            param_sharing: Whether gates in the same layer share a parameter. Defaults to False.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.cry([0, 1], torch.pi / 2)
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \ctrl[]{1} & \meter[2]{} & {} \\
                \lstick{} & \gate[1]{R_{y}(1.57)} & {} & {}
        """
        qubits_idx = self.register_idx(qubits_idx, 2)
        self.append(CRY(qubits_idx, param, param_sharing))

    def crz(self, qubits_idx: Union[Iterable[int], str] = 'cycle',
            param: Union[torch.Tensor, float] = None, param_sharing: bool = False) -> None:
        r"""Add controlled rotation gates about the z-axis.

        For a 2-qubit circuit, when `qubits_idx` is `[0, 1]`, the matrix form is:

        .. math::
            \mathit{CR_Z} =
            |0\rangle \langle 0|\otimes I + |1\rangle \langle 1|\otimes R_Z

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied. Defaults to 'cycle'.
            param: Parameters of the gates. Defaults to None.
            param_sharing: Whether gates in the same layer share a parameter. Defaults to False.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.crz([0, 1], torch.pi / 2)
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \ctrl[]{1} & \meter[2]{} & {} \\
                \lstick{} & \gate[1]{R_{z}(1.57)} & {} & {}
        """
        qubits_idx = self.register_idx(qubits_idx, 2)
        self.append(CRZ(qubits_idx, param, param_sharing))

    def cu(self, qubits_idx: Union[Iterable[int], str] = 'cycle',
           param: Union[torch.Tensor, float] = None, param_sharing: bool = False) -> None:
        r"""Add controlled single-qubit rotation gates.

        For a 2-qubit circuit, when `qubits_idx` is `[0, 1]`, the matrix form is given by a controlled-U gate.

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied. Defaults to 'cycle'.
            param: Parameters of the gate. Defaults to None.
            param_sharing: Whether gates in the same layer share a parameter. Defaults to False.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.cu([0, 1], torch.pi / 2)
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \ctrl[]{1} & \meter[2]{} & {} \\
                \lstick{} & \gate[1]{U(1.57, ...)} & {} & {}
        """
        qubits_idx = self.register_idx(qubits_idx, 2)
        self.append(CU(qubits_idx, param, param_sharing))

    def rxx(self, qubits_idx: Union[Iterable[int], str] = 'linear',
            param: Union[torch.Tensor, float] = None, param_sharing: bool = False) -> None:
        r"""Add RXX gates.

        The matrix form is:

        .. math::
            R_{XX}(\theta) =
            \begin{bmatrix}
                \cos\frac{\theta}{2} & 0 & 0 & -i\sin\frac{\theta}{2} \\
                0 & \cos\frac{\theta}{2} & -i\sin\frac{\theta}{2} & 0 \\
                0 & -i\sin\frac{\theta}{2} & \cos\frac{\theta}{2} & 0 \\
                -i\sin\frac{\theta}{2} & 0 & 0 & \cos\frac{\theta}{2}
            \end{bmatrix}

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied. Defaults to 'linear'.
            param: Parameters of the gate. Defaults to None.
            param_sharing: Whether gates in the same layer share a parameter. Defaults to False.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.rxx([0, 1], torch.pi / 2)
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[2]{R_{xx}(1.57)} & \meter[2]{} & {} \\
                \lstick{} & {} & {} & {}
        """
        qubits_idx = self.register_idx(qubits_idx, 2)
        self.append(RXX(qubits_idx, param, param_sharing))

    def ryy(self, qubits_idx: Union[Iterable[int], str] = 'linear',
            param: Union[torch.Tensor, float] = None, param_sharing: bool = False) -> None:
        r"""Add RYY gates.

        The matrix form is:

        .. math::
            R_{YY}(\theta) =
            \begin{bmatrix}
                \cos\frac{\theta}{2} & 0 & 0 & i\sin\frac{\theta}{2} \\
                0 & \cos\frac{\theta}{2} & -i\sin\frac{\theta}{2} & 0 \\
                0 & -i\sin\frac{\theta}{2} & \cos\frac{\theta}{2} & 0 \\
                i\sin\frac{\theta}{2} & 0 & 0 & \cos\frac{\theta}{2}
            \end{bmatrix}

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied. Defaults to 'linear'.
            param: Parameters of the gate. Defaults to None.
            param_sharing: Whether gates in the same layer share a parameter. Defaults to False.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.ryy([0, 1], torch.pi / 2)
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[2]{R_{yy}(1.57)} & \meter[2]{} & {} \\
                \lstick{} & {} & {} & {}
        """
        qubits_idx = self.register_idx(qubits_idx, 2)
        self.append(RYY(qubits_idx, param, param_sharing))

    def rzz(self, qubits_idx: Union[Iterable[int], str] = 'linear',
            param: Union[torch.Tensor, float] = None, param_sharing: bool = False) -> None:
        r"""Add RZZ gates.

        The matrix form is:

        .. math::
            R_{ZZ}(\theta) =
            \begin{bmatrix}
                e^{-i\theta/2} & 0 & 0 & 0 \\
                0 & e^{i\theta/2} & 0 & 0 \\
                0 & 0 & e^{i\theta/2} & 0 \\
                0 & 0 & 0 & e^{-i\theta/2}
            \end{bmatrix}

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied. Defaults to 'linear'.
            param: Parameters of the gate. Defaults to None.
            param_sharing: Whether gates in the same layer share a parameter. Defaults to False.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.rzz([0, 1], torch.pi / 2)
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[2]{R_{zz}(1.57)} & \meter[2]{} & {} \\
                \lstick{} & {} & {} & {}
        """
        qubits_idx = self.register_idx(qubits_idx, 2)
        self.append(RZZ(qubits_idx, param, param_sharing))

    def ms(self, qubits_idx: Union[Iterable[int], str] = 'cycle') -> None:
        r"""Add Mølmer-Sørensen (MS) gates.

        The matrix form is:

        .. math::
            \mathit{MS} = R_{XX}(-\pi/2) = \frac{1}{\sqrt{2}}
            \begin{bmatrix}
                1 & 0 & 0 & i \\
                0 & 1 & i & 0 \\
                0 & i & 1 & 0 \\
                i & 0 & 0 & 1
            \end{bmatrix}

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied. Defaults to 'cycle'.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.ms([0, 1])
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[2]{\text{MS}} & \meter[2]{} & {} \\
                \lstick{} & {} & {} & {}
        """
        qubits_idx = self.register_idx(qubits_idx, 2)
        self.append(MS(qubits_idx))

    def cswap(self, qubits_idx: Union[Iterable[int], str] = 'cycle') -> None:
        r"""Add CSWAP (Fredkin) gates.

        The matrix form is:

        .. math::
            \mathit{CSWAP} =
            \begin{bmatrix}
                1 & 0 & \cdots & 0 \\
                0 & 1 &         & 0 \\
                \vdots &  & \ddots  & \vdots \\
                0 & 0 & \cdots & 1
            \end{bmatrix}

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied. Defaults to 'cycle'.

        Examples:
            .. code-block:: python

                qc = Circuit(3)
                qc.cswap([0, 1, 2])
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \ctrl[]{1} & \meter[3]{} & {} \\
                \lstick{} & \gate[2,style={draw=gray, dashed}]{\permute{2,1}} & {} & {} \\
                \lstick{} & {} & {} & {}
        """
        qubits_idx = self.register_idx(qubits_idx, 2)
        self.append(CSWAP(qubits_idx))

    def ccx(self, qubits_idx: Union[Iterable[int], str] = 'cycle') -> None:
        r"""Add CCX (Toffoli) gates.

        The matrix form is:

        .. math::
            \mathit{CCX} =
            \begin{bmatrix}
                1 & 0 & \cdots & 0 \\
                0 & 1 &         & 0 \\
                \vdots &  & \ddots  & \vdots \\
                0 & 0 & \cdots & 1
            \end{bmatrix}

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied. Defaults to 'cycle'.

        Examples:
            .. code-block:: python

                qc = Circuit(3)
                qc.ccx([0, 1, 2])
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \ctrl[]{2} & \meter[3]{} & {} \\
                \lstick{} & \control{} & {} & {} \\
                \lstick{} & \targ{} & {} & {}
        """
        qubits_idx = self.register_idx(qubits_idx, 3)
        self.append(CCX(qubits_idx))

    def universal_two_qubits(self, qubits_idx: Union[List[int], str] = None,
                             param: Union[torch.Tensor, float] = None) -> None:
        r"""Add universal two-qubit gates. One such gate requires 15 parameters.

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied.
            param: Parameters of the gates. Defaults to None.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.universal_two_qubits([0, 1])
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                (The output latex code shows a universal 2-qubit layer.)
        """
        qubits_idx = self.register_idx(qubits_idx, None)
        self.append(Universal2(qubits_idx, param))

    def universal_three_qubits(self, qubits_idx: Optional[List[int]] = None,
                               param: Union[torch.Tensor, float] = None) -> None:
        r"""Add universal three-qubit gates. One such gate requires 81 parameters.

        Args:
            qubits_idx: Indices of the qubits on which the gates are applied.
            param: Parameters of the gates. Defaults to None.

        Examples:
            .. code-block:: python

                qc = Circuit(3)
                qc.universal_three_qubits([0, 1, 2])
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                (The output latex code shows a universal 3-qubit layer.)
        """
        qubits_idx = self.register_idx(qubits_idx, None)
        self.append(Universal3(qubits_idx, param))
    
    @_alias({"system_idx": "qubits_idx"})
    def universal_qudits(self, system_idx: List[int],
                         param: Union[torch.Tensor, float] = None, param_sharing: bool = False) -> None:
        r"""Add universal qudit gates. One such gate requires :math:`d^2 - 1` parameters,
        where :math:`d` is the gate dimension.

        Args:
            system_idx: Indices of the systems on which the gates are applied.
            param: Parameters of the gates. Defaults to None.
            param_sharing: Whether gates in the same layer share a parameter. Defaults to False.

        Examples:
            .. code-block:: python

                qc = Circuit(1, 3)
                qc.universal_qudits([0])
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[1]{\operatorname{UNI}_{3}} & \meter[1]{} & {}
        """
        system_idx = self.register_idx(system_idx, 1 if isinstance(system_idx, int) else len(system_idx))
        self.append(UniversalQudits(system_idx, [self.system_dim[idx] for idx in system_idx],
                                    param, param_sharing))
    
    @_alias({"system_idx": "qubits_idx"})
    def permute(self, perm: List[int], system_idx: List[int]) -> None:
        r"""Add a permutation gate.

        Args:
            perm: A list representing the permutation of subsystems.
            system_idx: Indices of the systems on which the gates are applied.

        Examples:
            .. code-block:: python

                qc = Circuit(3)
                qc.permute([1, 0, 2], [0, 1, 2])
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[3,style={draw=none}]{\permute{2,1,3}} & \meter[3]{} & {} \\
                \lstick{} & {} & {} & {} \\
                \lstick{} & {} & {} & {}
        """
        system_idx = self.register_idx(system_idx, len(system_idx))
        self.append(Permutation(perm, system_idx, [self.system_dim[idx] for idx in system_idx]))
        
    @_alias({"system_idx": "qubits_idx"})
    def control_permute(self, perm: List[int], system_idx: List[int], control_idx: int = -1) -> None:
        r"""Add a controlled permutation gate.

        Args:
            perm: A list representing the permutation of subsystems.
            system_idx: Indices of the systems on which the gates are applied.
            control_idx: the index that controls the oracle. Defaults to -1, meaning the highest index.

        Examples:
            .. code-block:: python

                qc = Circuit(4)
                qc.control_permute([1, 0, 2], [3, 0, 1, 2], 1)
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[3,style={draw=gray, dashed}]{\permute{2,1,3}} & \meter[4]{} & {} \\
                \lstick{} & {} & {} & {} \\
                \lstick{} & {} & {} & {} \\
                \lstick{} & \ctrl[]{-1} & {} & {}
        """
        if isinstance(system_idx[0], int):
            system_idx[0] = [system_idx[0]]
        _system_idx = system_idx[0] + system_idx[1:]
        _system_idx = self.register_idx(_system_idx, len(_system_idx))
        acted_system_dim = [self.system_dim[idx] for idx in _system_idx]

        ctrl_dim = math.prod(acted_system_dim[:len(system_idx[0])])
        if control_idx == -1:
            control_idx = ctrl_dim - 1
        else:
            assert control_idx < ctrl_dim, \
                f"Control index out of range: expect < {ctrl_dim}, got {control_idx}."
        
        permutation = utils.matrix._permutation(perm, acted_system_dim[len(system_idx[0]):])
        
        gate_info = {
            "name": "ctrl-permute",
            "api": "control_permute",
            "permute": perm,
            'plot_width': 0.2,
        }
        self.append(ControlOracle(
            permutation, system_idx, control_idx, acted_system_dim, gate_info))
    
    @_alias({"system_idx": "qubits_idx"})
    def oracle(self, oracle: torch.Tensor, system_idx: Union[List[Union[List[int], int]], int], 
               control_idx: Optional[int] = None, gate_name: Optional[str] = 'oracle',
               latex_name: Optional[str] = None) -> None:
        r"""Add an oracle gate.

        Args:
            oracle: Unitary oracle.
            system_idx: Indices of the systems on which the gate is applied.
            control_idx: the index that controls the oracle. Defaults to None.
            gate_name: name of the oracle.
            latex_name: LaTeX name of the gate. Defaults to gate_name.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.oracle(oracle=eye(2), system_idx=[0], latex_name=r'$Identity$')
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[1]{$Identity$}
        """
        gate_info = {
            'name': gate_name,
            'tex': r'\text{' + gate_name + r'}' if latex_name is None else latex_name
        }
        
        if control_idx is None:
            system_idx = [system_idx] if isinstance(system_idx, int) else system_idx
            system_idx = self.register_idx(system_idx, len(system_idx))
            
            acted_system_dim = [self.system_dim[idx] for idx in system_idx]
            self.append(Oracle(oracle, system_idx, acted_system_dim, gate_info))
        else:
            if isinstance(system_idx[0], int):
                system_idx[0] = [system_idx[0]]
            _system_idx = system_idx[0] + system_idx[1:]
            _system_idx = self.register_idx(_system_idx, len(_system_idx))
            acted_system_dim = [self.system_dim[idx] for idx in _system_idx]

            ctrl_dim = math.prod(acted_system_dim[:len(system_idx[0])])
            if control_idx == -1:
                control_idx = ctrl_dim - 1
            else:
                assert control_idx < ctrl_dim, \
                    f"Control index out of range: expect < {ctrl_dim}, got {control_idx}."
            self.append(ControlOracle(
                oracle, system_idx, control_idx, acted_system_dim, gate_info))

    @_alias({"system_idx": "qubits_idx"})
    def control_oracle(self, oracle: torch.Tensor, system_idx: List[Union[List[int], int]], control_idx: int = -1,
                        gate_name: Optional[str] = 'oracle', latex_name: Optional[str] = None) -> None:
        r"""Add a controlled oracle gate.

        Args:
            oracle: Unitary oracle.
            system_idx: Indices of the systems on which the gate is applied.
            control_idx: the index that controls the oracle. Defaults to -1.
            gate_name: name of the oracle.
            latex_name: LaTeX name of the gate.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.control_oracle(oracle=eye(2), system_idx=[0, 1], control_idx=0, latex_name=r'$Identity$')
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \octrl[]{1} & \meter[2]{} & {} \\
                \lstick{} & \gate[1]{$Identity$} & {} & {}
        """
        self.oracle(oracle, system_idx, control_idx, gate_name, latex_name)
        
    @_alias({"system_idx": "qubits_idx"})
    def param_oracle(self, generator: Callable[[torch.Tensor], torch.Tensor], num_acted_param: int,
                     system_idx: Union[List[int], int], control_idx: Optional[int] = None,
                     param: Union[torch.Tensor, float] = None, gate_name: Optional[str] = 'param-oracle',
                     latex_name: Optional[str] = None, support_batch: bool = False) -> None:
        r"""Add a parameterized oracle gate.

        Args:
            generator: Function to generate the oracle.
            num_acted_param: Number of parameters required.
            system_idx: Indices of the systems on which the gate acts.
            control_idx: The index that controls the oracle. Defaults to None.
            param: Input parameters for the gate. Defaults to None.
            gate_name: Name of the oracle.
            latex_name: LaTeX name of the gate.
            support_batch: Whether generator supports batched input.

        Examples:
            .. code-block:: python

                def rotation_generator(params: torch.Tensor) -> torch.Tensor:
                    theta = params[..., 0]
                    cos_theta = torch.cos(theta / 2).unsqueeze(-1)
                    sin_theta = torch.sin(theta / 2).unsqueeze(-1)
                    matrix = torch.cat([
                        torch.cat([cos_theta, -1j * sin_theta], dim=-1),
                        torch.cat([-1j * sin_theta, cos_theta], dim=-1)
                    ], dim=-2)
                    return matrix

                qc = Circuit(2)
                qc.param_oracle(
                    generator=rotation_generator,
                    num_acted_param=1,
                    system_idx=[0, 1],
                    control_idx=0,
                    param=None,
                    gate_name="ControlledRotation",
                    support_batch=True
                )
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \octrl[]{1} \\
                \lstick{} & \gate[1]{ControlledRotation(0.36)}
        """
        gate_info = {
            'name': gate_name,
            'tex': r'\text{' + gate_name + r'}' if latex_name is None else latex_name}
        
        if control_idx is None:
            system_idx = [system_idx] if isinstance(system_idx, int) else system_idx
            system_idx = self.register_idx(system_idx, len(system_idx))
            
            acted_system_dim = [self.system_dim[idx] for idx in system_idx]
            self.append(ParamOracle(
                generator, system_idx, param, num_acted_param, acted_system_dim, gate_info, support_batch))
        else:
            if isinstance(system_idx[0], int):
                system_idx[0] = [system_idx[0]]
            _system_idx = system_idx[0] + system_idx[1:]
            _system_idx = self.register_idx(_system_idx, len(_system_idx))
            acted_system_dim = [self.system_dim[idx] for idx in _system_idx]
            
            ctrl_dim = math.prod(acted_system_dim[:len(system_idx[0])])
            assert control_idx < ctrl_dim, f"Control index out of range: expect < {ctrl_dim}, got {control_idx}."
            self.append(ControlParamOracle(
                generator, system_idx, control_idx, param, num_acted_param, acted_system_dim, gate_info, support_batch))

    @_alias({"system_idx": "qubits_idx", "post_selection": "desired_result"})
    def measure(self, system_idx: Union[Iterable[int], int, str] = None, post_selection: Union[int, str] = None,
                if_print: bool = False, measure_basis: Optional[torch.Tensor] = None) -> None:
        r"""Perform a measurement on the specified systems.

        Args:
            system_idx: Systems to measure. Defaults to all.
            post_selection: The post-selection result. Defaults to None.
            if_print: Whether to print collapse info. Defaults to False.
            measure_basis: Measurement basis.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.measure()
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \meter[2]{} & {} \\
                \lstick{} & {} & {}
        """
        system_idx = self.register_idx(system_idx, None)
        acted_system_dim = [self.system_dim[idx] for idx in system_idx]
        self.append(Collapse(system_idx, acted_system_dim, post_selection, if_print, measure_basis))
    
    @_alias({"system_idx": "qubits_idx"})
    def locc(self, local_unitary: torch.Tensor, system_idx: List[Union[List[int], int]],
             label: str = 'M', latex_name: str = 'O') -> None:
        r"""Add a one-way LOCC protocol comprised of unitary operations.

        Args:
            local_unitary: The local unitary operation.
            system_idx: Systems on which the protocol is applied. The first element indicates the measure system.
            label: Label for measurement. Defaults to 'M'.
            latex_name: LaTeX name for the applied operator. Defaults to 'O'.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.locc(local_unitary=x(), system_idx=[0, 1])
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \meter{} & \push{M} \wireoverride{c} & \ctrl[vertical wire=c]{1}\wireoverride{c}  \\
                \lstick{} & {} & {} & \gate[1]{O^{(M)}}
        
        """
        if isinstance(system_idx[0], int):
            system_idx[0] = [system_idx[0]]
        measure_idx, apply_idx = system_idx[0], system_idx[1:]
        _system_idx = measure_idx + apply_idx
        _system_idx = self.register_idx(_system_idx, len(np.unique(_system_idx)))
        system_dim = [self.system_dim[idx] for idx in _system_idx]
        measure_dim, apply_dim = system_dim[:len(measure_idx)], system_dim[len(measure_idx):]
        
        gate = Oracle(local_unitary, apply_idx, apply_dim)
        self.append(OneWayLOCC(gate, measure_idx, measure_dim, label=label, latex_name=latex_name))
        
    @_alias({"system_idx": "qubits_idx"})
    def param_locc(self, generator: Callable[[torch.Tensor], torch.Tensor], num_acted_param: int, 
                   system_idx: List[Union[List[int], int]], param: Union[torch.Tensor, float] = None, 
                   label: str = 'M', latex_name: str = 'U', support_batch: bool = False) -> None:
        r"""Add a one-way LOCC protocol comprised of unitary operations, where the applied unitary is parameterized.

        Args:
            generator: Function to generate the oracle.
            num_acted_param: Number of parameters required.
            system_idx: Systems on which the protocol is applied. The first element indicates the measure system.
            param: Input parameters for the gate. Defaults to None.
            label: Label for measurement. Defaults to 'M'.
            latex_name: LaTeX name for the applied operator. Defaults to 'U'.
            support_batch: Whether generator supports batched input.

        """
        if isinstance(system_idx[0], int):
            system_idx[0] = [system_idx[0]]
        measure_idx, apply_idx = system_idx[0], system_idx[1:]
        _system_idx = measure_idx + apply_idx
        _system_idx = self.register_idx(_system_idx, len(np.unique(_system_idx)))
        system_dim = [self.system_dim[idx] for idx in _system_idx]
        measure_dim, apply_dim = system_dim[:len(measure_idx)], system_dim[len(measure_idx):]
        
        if param is None:
            float_dtype = intrinsic._get_float_dtype(self.dtype)
            expect_shape = intrinsic._format_param_shape([apply_idx], num_acted_param, param_sharing=False, batch_size=np.prod(measure_dim))
            param = torch.nn.Parameter(torch.rand(expect_shape, dtype=float_dtype) * 2 * np.pi)
        param_gate = ParamOracle(
            generator, apply_idx, param, num_acted_param, apply_dim, support_batch=support_batch)
        self.append(OneWayLOCC(param_gate, measure_idx, measure_dim, label=label, latex_name=latex_name))

    def linear_entangled_layer(self, qubits_idx: Optional[List[int]] = None, depth: int = 1,
                               param: Union[torch.Tensor, float] = None) -> None:
        r"""Add linear entangled layers consisting of Ry gates, Rz gates, and CNOT gates.

        Args:
            qubits_idx: Systems to apply the layer on. Defaults to all.
            depth: Number of layers. Defaults to 1.
            param: Parameters for the layer. Defaults to self-generated.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.linear_entangled_layer([0, 1], depth=4)
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                (The output latex code shows a linear entangled layer.)
        """
        qubits_idx = self.register_idx(qubits_idx, None)
        self.append(LinearEntangledLayer(qubits_idx, depth, param))

    def real_entangled_layer(self, qubits_idx: Optional[List[int]] = None, depth: int = 1,
                             param: Union[torch.Tensor, float] = None) -> None:
        r"""Add strongly entangled layers consisting of Ry gates and CNOT gates.

        Args:
            qubits_idx: Systems to apply the layer on. Defaults to all.
            depth: Number of layers. Defaults to 1.
            param: Layer parameters. Defaults to self-generated.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.real_entangled_layer([0, 1], depth=4)
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                (The output latex code shows a real entangled layer.)
        """
        qubits_idx = self.register_idx(qubits_idx, None)
        self.append(RealEntangledLayer(qubits_idx, depth, param))

    def complex_entangled_layer(self, qubits_idx: Optional[List[int]] = None, depth: int = 1,
                                param: Union[torch.Tensor, float] = None) -> None:
        r"""Add strongly entangled layers consisting of single-qubit rotation gates and CNOT gates.

        Args:
            qubits_idx: Systems to apply the layer on. Defaults to all.
            depth: Number of layers. Defaults to 1.
            param: Layer parameters. Defaults to self-generated.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.complex_entangled_layer([0, 1], depth=4)
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                (The output latex code shows a complex entangled layer.)
        """
        qubits_idx = self.register_idx(qubits_idx, None)
        self.append(ComplexEntangledLayer(qubits_idx, depth, param))

    def real_block_layer(self, qubits_idx: Optional[List[int]] = None, depth: int = 1,
                         param: Union[torch.Tensor, float] = None) -> None:
        r"""Add weakly entangled layers consisting of Ry gates and CNOT gates.

        Args:
            qubits_idx: Systems to apply the layer on. Defaults to all.
            depth: Number of layers. Defaults to 1.
            param: Layer parameters. Defaults to self-generated.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.real_block_layer([0, 1], depth=4)
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                (The output latex code shows a real block layer.)
        """
        qubits_idx = self.register_idx(qubits_idx, None)
        self.append(RealBlockLayer(qubits_idx, depth, param))

    def complex_block_layer(self, qubits_idx: Optional[List[int]] = None, depth: int = 1,
                            param: Union[torch.Tensor, float] = None) -> None:
        r"""Add weakly entangled layers consisting of single-qubit rotation gates and CNOT gates.

        Args:
            qubits_idx: Systems to apply the layer on. Defaults to all.
            depth: Number of layers. Defaults to 1.
            param: Layer parameters. Defaults to self-generated.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.complex_block_layer([0, 1], depth=4)
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                (The output latex code shows a complex block layer.)
        """
        qubits_idx = self.register_idx(qubits_idx, None)
        self.append(ComplexBlockLayer(qubits_idx, depth, param))

    def trotter(
        self, hamiltonian: Hamiltonian, time: float, qubits_idx: Optional[List[int]] = None, 
        num_steps: int = 1, order: int = 1, name: str = 'H'
    ) -> None:
        r"""Add Trotter decompositions of a Hamiltonian evolution operator.

        Args:
            hamiltonian: Hamiltonian of the system whose time evolution is to be simulated.
            time: Total evolution time.
            qubits_idx: Indices of the qubits on which the layer is applied. Defaults to ``None``.
            num_steps: Number of trotter blocks. Defaults to 1.
            order: Order of the Trotter-Suzuki decomposition. Defaults to 1.
            name: Name of the Hamiltonian. Defaults to 'H'.

        """
        qubits_idx = self.register_idx(qubits_idx, None)
        tau = time / num_steps
        self.append(TrotterLayer(hamiltonian, qubits_idx, tau, num_steps, order, name))

    def bit_flip(
            self, prob: Union[torch.Tensor, float],
            qubits_idx: Union[Iterable[int], int, str] = 'full'
    ) -> None:
        r"""Add bit flip channels.

        Args:
            prob: Probability of a bit flip.
            qubits_idx: Systems to apply the channel on. Defaults to 'full'.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.bit_flip(prob=0.5, qubits_idx=[0, 1])
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[1]{\mathcal{E}_{p = 0.5}^{\textrm{\tiny{(BF)}}}} \\
                \lstick{} & \gate[1]{\mathcal{E}_{p = 0.5}^{\textrm{\tiny{(BF)}}}}
        """
        qubits_idx = self.register_idx(qubits_idx, 1)
        self.append(BitFlip(prob, qubits_idx))

    def phase_flip(self, prob: Union[torch.Tensor, float],
                   qubits_idx: Union[Iterable[int], int, str] = 'full') -> None:
        r"""Add phase flip channels.

        Args:
            prob: Probability of a phase flip.
            qubits_idx: Systems to apply the channel on. Defaults to 'full'.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.phase_flip(prob=0.5, qubits_idx=[0, 1])
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[1]{\mathcal{E}_{p = 0.5}^{\textrm{\tiny{(PF)}}}} \\
                \lstick{} & \gate[1]{\mathcal{E}_{p = 0.5}^{\textrm{\tiny{(PF)}}}}
        """
        qubits_idx = self.register_idx(qubits_idx, 1)
        self.append(PhaseFlip(prob, qubits_idx))

    def bit_phase_flip(self, prob: Union[torch.Tensor, float],
                       qubits_idx: Union[Iterable[int], int, str] = 'full') -> None:
        r"""Add bit phase flip channels.

        Args:
            prob: Probability of a bit phase flip.
            qubits_idx: Systems to apply the channel on. Defaults to 'full'.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.bit_phase_flip(prob=0.5, qubits_idx=[0, 1])
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[1]{\mathcal{E}_{p = 0.5}^{\textrm{\tiny{(BPF)}}}} \\
                \lstick{} & \gate[1]{\mathcal{E}_{p = 0.5}^{\textrm{\tiny{(BPF)}}}}
        """
        qubits_idx = self.register_idx(qubits_idx, 1)
        self.append(BitPhaseFlip(prob, qubits_idx))

    def amplitude_damping(self, gamma: Union[torch.Tensor, float],
                          qubits_idx: Union[Iterable[int], int, str] = 'full') -> None:
        r"""Add amplitude damping channels.

        Args:
            gamma: Damping probability.
            qubits_idx: Systems to apply the damping on. Defaults to 'full'.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.amplitude_damping(gamma=0.5, qubits_idx=[0, 1])
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[1]{\mathcal{E}_{\gamma = 0.5}^{\textrm{\tiny{(AD)}}}} \\
                \lstick{} & \gate[1]{\mathcal{E}_{\gamma = 0.5}^{\textrm{\tiny{(AD)}}}}
        """
        qubits_idx = self.register_idx(qubits_idx, 1)
        self.append(AmplitudeDamping(gamma, qubits_idx))

    def generalized_amplitude_damping(self, gamma: Union[torch.Tensor, float],
                                      prob: Union[torch.Tensor, float],
                                      qubits_idx: Union[Iterable[int], int, str] = 'full') -> None:
        r"""Add generalized amplitude damping channels.

        Args:
            gamma: Damping probability.
            prob: Excitation probability.
            qubits_idx: Systems to apply the channel on. Defaults to 'full'.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.generalized_amplitude_damping(gamma=0.5, prob=0.5, qubits_idx=[0, 1])
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[1]{\mathcal{E}_{\gamma = 0.5, p = 0.5}^{\textrm{\tiny{(GAD)}}}} \\
                \lstick{} & \gate[1]{\mathcal{E}_{\gamma = 0.5, p = 0.5}^{\textrm{\tiny{(GAD)}}}}
        """
        qubits_idx = self.register_idx(qubits_idx, 1)
        self.append(GeneralizedAmplitudeDamping(gamma, prob, qubits_idx))

    def phase_damping(self, gamma: Union[torch.Tensor, float],
                      qubits_idx: Union[Iterable[int], int, str] = 'full') -> None:
        r"""Add phase damping channels.

        Args:
            gamma: Phase damping parameter.
            qubits_idx: Systems to apply the channel on. Defaults to 'full'.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.phase_damping(gamma=0.5, qubits_idx=[0, 1])
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[1]{\mathcal{E}_{\gamma = 0.5}^{\textrm{\tiny{(PD)}}}} \\
                \lstick{} & \gate[1]{\mathcal{E}_{\gamma = 0.5}^{\textrm{\tiny{(PD)}}}}
        """
        qubits_idx = self.register_idx(qubits_idx, 1)
        self.append(PhaseDamping(gamma, qubits_idx))

    def depolarizing(self, prob: Union[torch.Tensor, float],
                     qubits_idx: Union[Iterable[int], int, str] = 'full') -> None:
        r"""Add depolarizing channels.

        Args:
            prob: Depolarizing probability.
            qubits_idx: Systems to apply the channel on. Defaults to 'full'.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.depolarizing(prob=0.5, qubits_idx=[0, 1])
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[1]{\mathcal{D}_{p = 0.5}} \\
                \lstick{} & \gate[1]{\mathcal{D}_{p = 0.5}}
        """
        qubits_idx = self.register_idx(qubits_idx, 1)
        self.append(Depolarizing(prob, qubits_idx))

    def generalized_depolarizing(self, prob: Union[torch.Tensor, float],
                                  qubits_idx: Union[Iterable[int], int, str]) -> None:
        r"""Add a general depolarizing channel.

        Args:
            prob: Probabilities for the Pauli basis.
            qubits_idx: Systems to apply the channel on.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.generalized_depolarizing(prob=0.5, qubits_idx=[0, 1])
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[2]{\mathcal{D}_{p = 0.5}} \\
                \lstick{} & {}
        """
        qubits_idx = self.register_idx(qubits_idx, 1)
        self.append(GeneralizedDepolarizing(prob, qubits_idx))

    def pauli_channel(self, prob: Union[torch.Tensor, float],
                      qubits_idx: Union[Iterable[int], int, str] = 'full') -> None:
        r"""Add Pauli channels.

        Args:
            prob: Probabilities for the Pauli X, Y, and Z operators.
            qubits_idx: Systems to apply the channel on. Defaults to 'full'.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.pauli_channel(prob=[0.1, 0.3, 0.5])
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[1]{\mathcal{N}} \\
                \lstick{} & \gate[1]{\mathcal{N}}
        """
        qubits_idx = self.register_idx(qubits_idx, 1)
        self.append(PauliChannel(prob, qubits_idx))

    def reset_channel(self, prob: Union[torch.Tensor, float],
                      qubits_idx: Union[Iterable[int], int, str] = 'full') -> None:
        r"""Add reset channels.

        Args:
            prob: Probabilities for resetting to the basis states.
            qubits_idx: Systems to apply the channel on. Defaults to 'full'.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.reset_channel(prob=[0.5, 0.4], qubits_idx=[0, 1])
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[1]{\mathcal{N}} \\
                \lstick{} & \gate[1]{\mathcal{N}}
        """
        qubits_idx = self.register_idx(qubits_idx, 1)
        self.append(ResetChannel(prob, qubits_idx))

    def thermal_relaxation(self, const_t: Union[torch.Tensor, Iterable[float]], exec_time: Union[torch.Tensor, float],
                           qubits_idx: Union[Iterable[int], int, str] = 'full') -> None:
        r"""Add thermal relaxation channels.

        Args:
            const_t: The T1 and T2 relaxation times.
            exec_time: Gate execution time.
            qubits_idx: Systems to apply the channel on. Defaults to 'full'.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.thermal_relaxation(const_t=[600, 300], exec_time=500, qubits_idx=[0, 1])
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[1]{\mathcal{N}} \\
                \lstick{} & \gate[1]{\mathcal{N}}
        """
        qubits_idx = self.register_idx(qubits_idx, 1)
        self.append(ThermalRelaxation(const_t, exec_time, qubits_idx))

    def choi_channel(self, choi_repr: Iterable[torch.Tensor],
                     system_idx: Union[Iterable[Iterable[int]], Iterable[int], int]) -> None:
        r"""Add custom channels in the Choi representation.

        Args:
            choi_repr: Choi representation.
            system_idx: Systems to apply the channel on.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                X = torch.tensor(x(), dtype=torch.complex64)
                choi = X.kron(X) / 2
                qc.choi_channel(choi_repr=choi, system_idx=[0])
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[1]{\mathcal{N}}
        """
        system_idx = [system_idx] if isinstance(system_idx, int) else system_idx
        system_idx = self.register_idx(system_idx, len(system_idx))
        acted_system_dim = [self.system_dim[idx] for idx in system_idx]
        self.append(ChoiRepr(choi_repr, system_idx, acted_system_dim))

    def kraus_channel(self, kraus_oper: Iterable[torch.Tensor],
                      system_idx: Union[Iterable[Iterable[int]], Iterable[int], int]) -> None:
        r"""Add custom channels in the Kraus representation.

        Args:
            kraus_oper: Kraus operators.
            system_idx: Systems to apply the channel on.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.kraus_channel(kraus_oper=eye(2), system_idx=[0])
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[1]{\mathcal{N}}
        """
        system_idx = [system_idx] if isinstance(system_idx, int) else system_idx
        system_idx = self.register_idx(system_idx, len(system_idx))
        acted_system_dim = [self.system_dim[idx] for idx in system_idx]
        self.append(KrausRepr(kraus_oper, system_idx, acted_system_dim))

    def stinespring_channel(self, stinespring_repr: Iterable[torch.Tensor],
                             system_idx: Union[Iterable[Iterable[int]], Iterable[int], int]) -> None:
        r"""Add custom channels in the Stinespring representation.

        Args:
            stinespring_repr: Stinespring representation.
            system_idx: Systems to apply the channel on.

        Examples:
            .. code-block:: python

                qc = Circuit(2)
                qc.stinespring_channel(stinespring_repr=eye(2), system_idx=[0])
                print(f'The latex code of this circuit is:\n{qc.to_latex()}')

            ::

                The latex code of this circuit is:
                \lstick{} & \gate[1]{\mathcal{N}}
        """
        system_idx = [system_idx] if isinstance(system_idx, int) else system_idx
        system_idx = self.register_idx(system_idx, len(system_idx))
        acted_system_dim = [self.system_dim[idx] for idx in system_idx]
        self.append(StinespringRepr(stinespring_repr, system_idx, acted_system_dim))



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The source file of the class for quantum circuit templates.
"""

from typing import List, Optional

import torch

from ..operator import CNOT, RY, RZ, U3, H, ParamGate, S
from .container import Layer

__all__ = ['LinearEntangledLayer', 'RealEntangledLayer', 'ComplexEntangledLayer', 'ComplexBlockLayer', 'RealBlockLayer', 'Universal2', 'Universal3']

class LinearEntangledLayer(Layer):
    r"""Linear entangled layers consisting of Ry gates, Rz gates, and CNOT gates.

    Args:
        qubits_idx: Indices of the qubits on which the layer is applied.
        depth: Number of layers.
        param: Initial parameters for the layer. Defaults to be self-generated.
    
    """
    def __init__(self, qubits_idx: List[int], depth: int,
                 param: Optional[torch.Tensor] = None) -> None:
        super().__init__(qubits_idx, depth, 'Linear Entangled Layer')
        self._assert_qubits()
        
        param_shape = [depth, 2, len(qubits_idx)]
        param = self._format_layer_param(param, param_shape)
        self.__add_layer(param, self.system_idx, self.num_systems)
    
    def __add_layer(self, list_param: torch.Tensor, qubits_idx: List[int], num_qubits: int) -> None:
        acted_list = [(qubits_idx[idx], qubits_idx[idx + 1])
                      for idx in range(num_qubits - 1)]
        
        for param in list_param:
            self.append(RY(qubits_idx, param=torch.nn.Parameter(param[0].view([-1, 1, 1]))))
            self.append(CNOT(qubits_idx=acted_list))
            self.append(RZ(qubits_idx, param=torch.nn.Parameter(param[1].view([-1, 1, 1]))))
            self.append(CNOT(qubits_idx=acted_list))


class RealEntangledLayer(Layer):
    r"""Strongly entangled layers consisting of Ry gates and CNOT gates.

    Args:
        qubits_idx: Indices of the qubits on which the layer is applied.
        depth: Number of layers. Defaults to ``1``.
        param: Initial parameters for the layer. Defaults to be self-generated.

    Note:
        The mathematical representation of this layer of quantum gates is a real unitary matrix.
        This ansatz is from the following paper: https://arxiv.org/pdf/1905.10876.pdf.
    """
    def __init__(self, qubits_idx: List[int], depth: int,
                 param: Optional[torch.Tensor] = None) -> None:
        super().__init__(qubits_idx, depth, 'Real Entangled Layer')
        self._assert_qubits()
        
        param_shape = [depth, len(qubits_idx)]
        param = self._format_layer_param(param, param_shape)
        self.__add_layer(param, self.system_idx, self.num_systems)
    
    def __add_layer(self, list_param: torch.Tensor, qubits_idx: List[int], num_qubits: int) -> None:
        acted_list = [(qubits_idx[idx], qubits_idx[(idx + 1) % num_qubits])
                      for idx in range(num_qubits)]

        for param in list_param:
            self.append(RY(qubits_idx, param=torch.nn.Parameter(param.view([-1, 1, 1]))))
            self.append(CNOT(qubits_idx=acted_list))
            

class ComplexEntangledLayer(Layer):
    r"""Strongly entangled layers consisting of single-qubit rotation gates and CNOT gates.

    Args:
        qubits_idx: Indices of the qubits on which the layer is applied.
        depth: Number of layers. Defaults to ``1``.
        param: Initial parameters for the layer. Defaults to be self-generated.

    Note:
        The mathematical representation of this layer of quantum gates is a complex unitary matrix.
        This ansatz is from the following paper: https://arxiv.org/abs/1804.00633.
    """
    def __init__(self, qubits_idx: List[int], depth: int,
                 param: Optional[torch.Tensor] = None) -> None:
        super().__init__(qubits_idx, depth, 'Complex Entangled Layer')
        self._assert_qubits()
        
        param_shape = [depth, len(qubits_idx), 3]
        param = self._format_layer_param(param, param_shape)
        self.__add_layer(param, self.system_idx, self.num_systems)
    
    def __add_layer(self, list_param: torch.Tensor, qubits_idx: List[int], num_qubits: int) -> None:
        acted_list = [(qubits_idx[idx], qubits_idx[(idx + 1) % num_qubits])
                      for idx in range(num_qubits)]

        for param in list_param:
            self.append(U3(qubits_idx, param=torch.nn.Parameter(param.view([-1, 1, 3]))))
            self.append(CNOT(qubits_idx=acted_list))


class RealBlockLayer(Layer):
    r"""Weakly entangled layers consisting of CNOT gates surrounded by RY gates

    Args:
        qubits_idx: Indices of the qubits on which the layer is applied.
        depth: Number of layers. Defaults to ``1``.
        param: Initial parameters for the layer. Defaults to be self-generated.

    Note:
        The mathematical representation of this layer of quantum gates is a real unitary matrix.
    
    """
    def __init__(self, qubits_idx: List[int], depth: int,
                 param: Optional[torch.Tensor] = None) -> None:
        super().__init__(qubits_idx, depth, 'Real Block Layer')
        self._assert_qubits()
        
        num_qubits = self.num_systems
        param_shape = [2 * depth * (num_qubits - 1) + num_qubits]
        param = self._format_layer_param(param, param_shape)
        self.__add_layer(param, self.system_idx, num_qubits, depth)
    
    def __add_layer(self, list_param: torch.Tensor, qubits_idx: List[int], num_qubits: int, depth: int) -> None:
        list_param, end_param = list_param[:-num_qubits], list_param[-num_qubits:]
        for param in list_param.view([depth, -1]):
            param_0, param_1 = param[:num_qubits], param[num_qubits:]
            self.__add_ry_layer(param_0, qubits_idx, [0, num_qubits - 1])
            
            if num_qubits % 2 == 0:
                self.__add_cnot_layer(qubits_idx, [0, num_qubits - 1])
                if num_qubits > 2:
                    self.__add_ry_layer(param_1, qubits_idx, [1, num_qubits - 2])
                    self.__add_cnot_layer(qubits_idx, [1, num_qubits - 2])
            else:
                self.__add_cnot_layer(qubits_idx, [0, num_qubits - 2])
                self.__add_ry_layer(param_1, qubits_idx, [1, num_qubits - 2])
                self.__add_cnot_layer(qubits_idx, [1, num_qubits - 1])
        
        self.__add_ry_layer(end_param, qubits_idx, [0, num_qubits - 1])

    def __add_cnot_layer(self, qubits_idx: List[int], position: List[int]) -> None:
        cnot_acted_list = [[qubits_idx[i], qubits_idx[i+1]] for i in range(position[0], position[1], 2)]
        self.append(CNOT(cnot_acted_list))

    def __add_ry_layer(self, param: torch.Tensor, qubits_idx: List[int], position: List[int]):
        ry_acted_list = list(range(qubits_idx[position[0]], qubits_idx[position[1]] + 1))
        self.append(RY(ry_acted_list, param=torch.nn.Parameter(param.view([-1, 1, 1]))))


class ComplexBlockLayer(Layer):
    r"""Weakly entangled layers consisting of CNOT gates surrounded by U3 gates

    Args:
        qubits_idx: Indices of the qubits on which the layer is applied.
        depth: Number of layers. Defaults to ``1``.
        param: Initial parameters for the layer. Defaults to be self-generated.

    Note:
        The mathematical representation of this layer of quantum gates is a complex unitary matrix.
    
    """
    def __init__(self, qubits_idx: List[int], depth: int,
                 param: Optional[torch.Tensor] = None) -> None:
        super().__init__(qubits_idx, depth, 'Complex Block Layer')
        self._assert_qubits()
        
        num_qubits = self.num_systems
        param_shape = [3 * (2 * depth * (num_qubits - 1) + num_qubits)]
        param = self._format_layer_param(param, param_shape)
        self.__add_layer(param, self.system_idx, num_qubits, depth)
    
    def __add_layer(self, list_param: torch.Tensor, qubits_idx: List[int], num_qubits: int, depth: int) -> None:
        list_param, end_param = list_param[:-num_qubits * 3], list_param[-num_qubits * 3:]
        for param in list_param.view([depth, -1, 3]):
            param_0, param_1 = param[:num_qubits], param[num_qubits:]
            self.__add_u3_layer(param_0, qubits_idx, [0, num_qubits - 1])
            
            if num_qubits % 2 == 0:
                self.__add_cnot_layer(qubits_idx, [0, num_qubits - 1])
                if num_qubits > 2:
                    self.__add_u3_layer(param_1, qubits_idx, [1, num_qubits - 2])
                    self.__add_cnot_layer(qubits_idx, [1, num_qubits - 2])
            else:
                self.__add_cnot_layer(qubits_idx, [0, num_qubits - 2])
                self.__add_u3_layer(param_1, qubits_idx, [1, num_qubits - 2])
                self.__add_cnot_layer(qubits_idx, [1, num_qubits - 1])
        
        self.__add_u3_layer(end_param, qubits_idx, [0, num_qubits - 1])

    def __add_cnot_layer(self, qubits_idx: List[int], position: List[int]) -> None:
        cnot_acted_list = [[qubits_idx[i], qubits_idx[i+1]] for i in range(position[0], position[1], 2)]
        self.append(CNOT(cnot_acted_list))

    def __add_u3_layer(self, param: torch.Tensor, qubits_idx: List[int], position: List[int]):
        u3_acted_list = list(range(qubits_idx[position[0]], qubits_idx[position[1]] + 1))
        self.append(U3(u3_acted_list, param=torch.nn.Parameter(param.view([-1, 1, 3]))))


class Universal2(Layer):
    r"""A circuit layer representing universal two-qubit gates. One of such a layer requires 15 parameters.
    
    Args:
        qubits_idx: Indices of the qubits on which the layer is applied.
        param: Initial parameters for the layer. Defaults to be self-generated.
    
    """
    def __init__(self, qubits_idx: List[int], param: Optional[torch.Tensor] = None) -> None:
        super().__init__(qubits_idx, 1, 'Universal 2-qubit Layer')
        self._assert_qubits()
        
        assert len(qubits_idx) == 2, \
            f"The width of the layer should be 2: received indices {qubits_idx}."
        
        param = self._format_layer_param(param, [15])
        self.__add_layer(param, self.system_idx)
    
    def __add_layer(self, theta: torch.Tensor, qubits_idx: List[int]) -> None:
        u3_param_1, u3_param_2 = theta[:6], theta[9:]
        
        self.append(U3([qubits_idx[0], qubits_idx[1]], param=torch.nn.Parameter(u3_param_1.view([-1, 1, 3]))))
        self.append(CNOT(qubits_idx=[[qubits_idx[1], qubits_idx[0]]]))
        
        self.append(RZ(qubits_idx[0], param=torch.nn.Parameter(theta[6].view([-1, 1, 1]))))
        self.append(RY(qubits_idx[1], param=torch.nn.Parameter(theta[7].view([-1, 1, 1]))))
        self.append(CNOT(qubits_idx=[[qubits_idx[0], qubits_idx[1]]]))
        
        self.append(RY(qubits_idx[1], param=torch.nn.Parameter(theta[8].view([-1, 1, 1]))))
        self.append(CNOT(qubits_idx=[[qubits_idx[1], qubits_idx[0]]]))
        self.append(U3([qubits_idx[0], qubits_idx[1]], param=torch.nn.Parameter(u3_param_2.view([-1, 1, 3]))))


class Universal3(Layer):
    r"""A circuit layer representing universal three-qubit gates. One of such a layer requires 81 parameters.
    
    Args:
        qubits_idx: Indices of the qubits on which the layer is applied.
        param: Initial parameters for the layer. Defaults to be self-generated.
    
    """
    def __init__(self, qubits_idx: List[int], param: Optional[torch.Tensor] = None) -> None:
        super().__init__(qubits_idx, 1, 'Universal 3-qubit Layer')
        self._assert_qubits()
        
        assert len(qubits_idx) == 3, \
            f"The width of the layer should be 3: received indices {qubits_idx}."
        
        param = self._format_layer_param(param, [81])
        
        self.__add_layer(param, self.system_idx)
        
    @property
    def param(self) -> torch.Tensor:
        # This overload is for alignment with the original implementation.
        psi, phi = torch.tensor([]), torch.tensor([])
        switch_to_psi = True
        current_psi_count, current_phi_count = 0, 0
        for op in self.children():
            if isinstance(op, ParamGate):
                param = torch.flatten(op.theta)
                
                if switch_to_psi:
                    current_psi_count += len(param)
                    psi = torch.cat([psi, param])
                    
                    if current_psi_count % 15 == 0:
                        switch_to_psi = False
                        continue
                
                else:
                    current_phi_count += len(param)
                    phi = torch.cat([phi, param])
                    
                    if current_phi_count % 6 == 0:
                        switch_to_psi = True
                        continue
        return torch.cat([psi, phi])
        
    def __add_layer(self, theta: torch.Tensor, qubits_idx: List[int]) -> None:
        psi, phi = theta[:60].view([4, 15]), theta[60:].view([7, 3])
        
        self.extend(Universal2([qubits_idx[0], qubits_idx[1]], param=psi[0]))
        self.append(U3(qubits_idx[2], param=torch.nn.Parameter(phi[0].view([-1, 1, 3]))))
        
        self.__block_u(phi[1], qubits_idx)
        
        self.extend(Universal2([qubits_idx[0], qubits_idx[1]], param=psi[1]))
        self.append(U3(qubits_idx[2], param=torch.nn.Parameter(phi[2].view([-1, 1, 3]))))
        
        self.__block_v(phi[3], qubits_idx)
        
        self.extend(Universal2([qubits_idx[0], qubits_idx[1]], param=psi[2]))
        self.append(U3(qubits_idx[2], param=torch.nn.Parameter(phi[4].view([-1, 1, 3]))))
        
        self.__block_u(phi[5], qubits_idx)
        
        self.extend(Universal2([qubits_idx[0], qubits_idx[1]], param=psi[3]))
        self.append(U3(qubits_idx[2], param=torch.nn.Parameter(phi[6].view([-1, 1, 3]))))
        
    def __block_u(self, phi: torch.Tensor, qubits_idx: List[int]) -> None:
        self.append(CNOT([[qubits_idx[1], qubits_idx[2]]]))
        self.append(RY(qubits_idx[1], param=torch.nn.Parameter(phi[0].view([-1, 1, 1]))))
        
        self.append(CNOT([[qubits_idx[0], qubits_idx[1]]]))
        self.append(RY(qubits_idx[1], param=torch.nn.Parameter(phi[1].view([-1, 1, 1]))))
        
        self.append(CNOT([[qubits_idx[0], qubits_idx[1]],
                          [qubits_idx[1], qubits_idx[2]]]))
        self.append(H(qubits_idx[2]))
        
        self.append(CNOT([[qubits_idx[1], qubits_idx[0]],
                          [qubits_idx[0], qubits_idx[2]],
                          [qubits_idx[1], qubits_idx[2]]]))
        self.append(RZ(qubits_idx[2], param=torch.nn.Parameter(phi[2].view([-1, 1, 1]))))
        self.append(CNOT([[qubits_idx[1], qubits_idx[2]],
                          [qubits_idx[0], qubits_idx[2]]]))
        
    def __block_v(self, phi: torch.Tensor, qubits_idx: List[int]) -> None:
        self.append(CNOT([[qubits_idx[2], qubits_idx[0]],
                          [qubits_idx[1], qubits_idx[2]],
                          [qubits_idx[2], qubits_idx[1]]]))
        
        self.append(RY(qubits_idx[2], param=torch.nn.Parameter(phi[0].view([-1, 1, 1]))))
        self.append(CNOT([[qubits_idx[1], qubits_idx[2]]]))
        self.append(RY(qubits_idx[2], param=torch.nn.Parameter(phi[1].view([-1, 1, 1]))))
        self.append(CNOT([[qubits_idx[1], qubits_idx[2]]]))
        
        self.append(S(qubits_idx[2]))
        self.append(CNOT([[qubits_idx[2], qubits_idx[0]],
                          [qubits_idx[0], qubits_idx[1]],
                          [qubits_idx[1], qubits_idx[0]]]))
        
        self.append(H(qubits_idx[2]))
        self.append(CNOT([[qubits_idx[0], qubits_idx[2]]]))
        self.append(RZ(qubits_idx[2], param=torch.nn.Parameter(phi[2].view([-1, 1, 1]))))
        self.append(CNOT([[qubits_idx[0], qubits_idx[2]]]))



# !/usr/bin/env python3
# Copyright (c) 2025 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The source file of the Sequential class.
"""

import itertools
import warnings
from typing import Dict, Generator, Iterable, List, Optional, Tuple, Union

import numpy as np
import torch
from torch.nn.parameter import Parameter

from ..core import (Operator, OperatorInfoType, State, get_backend, get_dtype,
                    get_float_dtype)
from ..core.intrinsic import (_display_png, _format_sequential_idx,
                              _replace_indices)
from ..core.latex import OperatorListDrawer
from ..database import std_basis, zero_state


class OperatorList(torch.nn.Sequential):
    r"""Sequential container for QuAIRKit operators.

    Args:
        num_systems: number of systems in the operator list.
        system_dim: dimension of systems of this list. Can be a list of system dimensions 
            or an int representing the dimension of all systems.
        physical_idx: physical indices of systems.
    
    """
    def __init__(self, num_systems: Optional[int], 
                 system_dim: Union[List[int], int],
                 physical_idx: Optional[List[int]]) -> None:
        self.__register_list(num_systems, system_dim, physical_idx)

        super().__init__()
        self.backend = get_backend()
        self.dtype = get_dtype()
    
    def __getitem__(self, key) -> Union[Operator, 'OperatorList']:
        return super().__getitem__(key)
    
    def __register_list(self, num_systems: Optional[int], 
                        system_dim: Union[List[int], int],
                        physical_idx: Optional[List[int]] = None) -> None:
        r"""Register the operator list with input information.
        """
        if isinstance(system_dim, int):
            assert num_systems, \
                ("Since QuAIRKit v0.4.0, system information need to be specified during circuit initialization. " + 
                 "One can change the number of systems by attribute `num_systems`.")
            system_dim = [system_dim] * num_systems
                
        elif num_systems is None:
            num_systems = len(system_dim)
        
        else:
            assert num_systems == len(system_dim), \
                f"num_systems and system_dim do not agree: received {num_systems} and {system_dim}"
        
        self.__system_dim: List[int] = system_dim
        
        if physical_idx is None:
            physical_idx = list(range(num_systems))
        else:
            assert len(set(physical_idx)) == len(physical_idx) == num_systems, \
                f"Duplicate or too less system indices: received {physical_idx}, expected of length {num_systems}"
        self.__system_idx: List[int] = physical_idx # physical index of systems

    @property
    def num_qubits(self) -> int:
        r"""Number of qubits in this circuit.
        """
        return 0 if isinstance(self.__system_dim, int) else self.__system_dim.count(2)
    
    @property
    def num_qutrits(self) -> int:
        r"""Number of qutrits in this circuit.
        """
        return 0 if isinstance(self.__system_dim, int) else self.__system_dim.count(3)

    @property
    def system_dim(self) -> List[int]:
        r"""Dimension of systems in this circuit.
        """
        return self.__system_dim.copy()
    
    @property
    def equal_dim(self) -> bool:
        r"""Whether the systems in this circuit have the same dimension.
        """
        return len(set(self.__system_dim)) == 1
    
    @property
    def system_idx(self) -> List[int]:
        r"""List of physical indices of systems.
        """
        return self.__system_idx.copy()
    
    @system_idx.setter
    def system_idx(self, system_idx: List[int]) -> None:
        r"""Set the physical indices of systems.
        
        Args:
            system_idx: physical indices of systems.
        
        """
        assert (
            len(set(system_idx)) == len(system_idx) == self.num_systems
        ), f"Duplicate or too less system indices: received {system_idx}"

        new_map = {physical: system_idx[logical] for logical, physical in enumerate(self.__system_idx)}
        for op in self.children():
            op.system_idx = _replace_indices(op.system_idx, new_map)
        self.__system_idx = system_idx
    
    @property
    def num_systems(self) -> int:
        r"""Number of logical systems.
        """
        return len(self.__system_idx)
    
    def add_systems(self, num_new_systems: int, 
                    new_system_dim: Optional[Union[int, List[int]]] = None, 
                    new_physical_idx: Optional[List[int]] = None) -> None:
        r"""Add new systems to the list.
        
        Args:
            num_new_systems: number of new systems to be added.
            new_system_dim: dimension of new systems. Defaults to be the same as other systems.
            new_physical_idx: physical indices of new systems. Defaults to start from the largest index.
        """
        if new_system_dim is None:
            assert self.equal_dim, \
                "Need to specify dimensions of new systems when circuit systems have different dimensions"
        elif isinstance(new_system_dim, int):
            new_system_dim = [new_system_dim] * num_new_systems
        else:
            assert len(new_system_dim) == num_new_systems, \
                f"Wrong format of new_system_dim: received {new_system_dim}, expected of length {num_new_systems}"
        
        num_old_systems = self.num_systems
        if new_physical_idx is None:
            new_physical_idx = list(range(num_old_systems, num_old_systems + num_new_systems))
        else:
            assert len(set(new_physical_idx)) == len(new_physical_idx) == num_new_systems, \
                f"Duplicate or too less system indices: received {new_physical_idx}"
            assert set(new_physical_idx).isdisjoint(self.__system_idx), \
                f"Duplicate physical system indices: received {new_physical_idx}, existing {self.__system_idx}"
        
        self.__system_idx = self.__system_idx + new_physical_idx
        self.__system_dim = self.__system_dim + new_system_dim
    
    def sort(self) -> None:
        r"""Sort the systems in the circuit by their physical indices.
        """
        for op in self.children():
            if isinstance(op, OperatorList):
                op.sort()
        
        _system_idx, _system_dim = zip(*sorted(zip(self.system_idx, self.system_dim)))
        self.__system_idx, self.__system_dim = list(_system_idx), list(_system_dim)
    
    @num_systems.setter
    def num_systems(self, num_systems: int) -> None:
        r"""Set the number of logical systems.
        
        Args:
            num_new_systems: number of total systems.
        
        """
        if (old_num_systems := self.num_systems) == num_systems:
            return
        assert num_systems >= old_num_systems, \
            f"Incorrect number of systems: received {num_systems}, expected >= {old_num_systems}"
        
        self.add_systems(num_systems - old_num_systems)

    def register_idx(self, operator_idx: Optional[Union[Iterable[int], int, str]], num_acted_system: int) -> List[List[int]]:
        r"""Update sequential according to input operator index information, or report error.

        Args:
            operator_idx: input system indices of the operator. None means acting on all systems.
            num_acted_system: number of systems that one operator acts on.
        
        Returns:
            the formatted system indices.
        
        """
        if operator_idx is None or isinstance(operator_idx, str):
            assert self.equal_dim, \
                f"The sequential's systems have different dimensions. Invalid input qubit idx: {operator_idx}"
        if operator_idx is None:
            return self.system_idx

        num_systems = self.num_systems
        operator_idx = _format_sequential_idx(operator_idx, num_systems, num_acted_system)

        assert (
            (max_idx := np.max(operator_idx)) < num_systems
        ), (f"Invalid input system idx: {max_idx} cannot match with {num_systems} systems." + 
            " One may change the number of systems by calling property `num_systems` or method `add_systems`.")

        return _replace_indices(operator_idx, self.__system_idx)

    def append(self, op: Union[Operator, 'OperatorList']) -> 'OperatorList':
        r"""Appends an operator or an operator sub-list to the end.

        Args:
            op: module to append
        
        """
        if isinstance(op, (Operator, OperatorList)):
            return super().append(op)
        elif isinstance(op, torch.nn.Module):
            warnings.warn(
                "OperatorList shall only append quantum operator: received normal module", UserWarning)
        else:
            warnings.warn(
                f"Unrecognized input: received {type(op)}", UserWarning)
        return super().append(op)
            
    def __iadd__(self, other: 'OperatorList') -> 'OperatorList':
        return self.extend(other)

    def extend(self, sequential: 'OperatorList') -> 'OperatorList':
        r"""Extend the list with another sequential

        Args:
            sequential: a sequential of operators to be extended

        Returns:
            Concatenation of two quantum operator sequential
        
        """
        new_idx, new_dim = [], []
        for idx, physical_idx in enumerate(sequential.system_idx):
            if physical_idx in self.system_idx:
                physical_dim = sequential.system_dim[idx]
                expected_dim = self.__system_dim[self.system_idx.index(physical_idx)]
                assert physical_dim == expected_dim, \
                    f"Physical system {physical_idx}: received dim {physical_dim}, expected {expected_dim}"
            else:
                new_idx.append(physical_idx)
                new_dim.append(sequential.system_dim[idx])
        
        self.__system_idx = self.__system_idx + new_idx
        self.__system_dim = self.__system_dim + new_dim
        
        return super().extend(sequential)
    
    def operators(self) -> Generator[Operator, None, None]:
        r"""Yield all operators in this list.

        Returns:
            A generator of operators
        
        """
        for op in self:
            if isinstance(op, OperatorList):
                yield from op.operators()
            else:
                yield op

    def to(self, dtype: Optional[str] = None):
        if dtype is not None:
            self.dtype = dtype

        for operator in self:
            operator.to(backend=self.backend, dtype=self.dtype)

    @property
    def operator_history(self) -> List[Union[OperatorInfoType, List[OperatorInfoType]]]:
        r"""Return the operator history of this Sequential
        """
        operator_history = []
        operator_history.extend(
            op.operator_history if isinstance(op, OperatorList) else op.info
            for op in self.children()
        )
        return operator_history

    @property
    def param(self) -> torch.Tensor:
        r"""Flattened parameters in this list.
        """
        assert self._modules, \
                    "The operator list is empty, please add some operators first."

        list_params = []
        for op in self.children():
            if isinstance(op, OperatorList):
                list_params.append(op.param)
                continue
            list_params.extend([torch.flatten(param.detach()) for param in op.parameters()])
        
        return torch.cat(list_params) if list_params else torch.tensor([])

    @property
    def grad(self) -> np.ndarray:
        r"""Gradients with respect to the flattened parameters.
        """
        assert self._modules, \
            "The operator list is empty, please add some operators first."
        grad_list = []
        for param in self.parameters():
            assert param.grad is not None, (
                'The gradient is None, run the backward first before calling this property, '
                'otherwise check where the gradient chain is broken.')
            grad_list.append(param.grad.detach().numpy().flatten())
        return np.concatenate(grad_list) if grad_list != [] else grad_list

    def update_param(self, theta: Union[torch.Tensor, np.ndarray, float], 
                     idx: Optional[Union[int, None]] = None) -> None:
        r"""Replace parameters of all/one layer(s) by ``theta``.

        Args:
            theta: New parameters
            idx: Index of replacement. Defaults to None, referring to all layers.
        """
        if not isinstance(theta, torch.Tensor):
            theta = torch.tensor(theta)
        theta = theta.reshape([-1]).to(dtype=get_float_dtype())

        if idx is None:
            assert self.param.shape == theta.shape, \
                f"the shape of input parameters is not correct: expect {self.param.shape}, received {theta.shape}"
            for layer in self.children():
                if isinstance(layer, OperatorList):
                    num_param = int(torch.numel(layer.param))
                    layer.update_param(theta[:num_param], None)
                    continue
                
                for name, param in layer.named_parameters():
                    num_param = int(torch.numel(param))
                    layer.register_parameter(name, Parameter(theta[:num_param].reshape(param.shape)))

                    if num_param == theta.shape[0]:
                        return
                    theta = theta[num_param:]
        
        elif isinstance(idx, int):
            layer = next(itertools.islice(self.children(), idx, None))
            if isinstance(layer, OperatorList):
                layer.update_param(theta, None)
                return
            
            assert theta.shape == torch.cat([torch.flatten(param) for param in layer.parameters()]).shape, (
                "The shape of input parameters is not correct.")

            for name, param in layer.named_parameters():
                num_param = int(torch.numel(param))
                layer.register_parameter(name, Parameter(theta[:num_param].reshape(param.shape)))

                if num_param == theta.shape[0]:
                    return
                theta = theta[num_param:]
        else:
            raise ValueError("idx must be an integer or None")

    def transfer_static(self) -> None:
        r"""
        set ``stop_gradient`` of all parameters of the list as ``True``
        """
        for layer in self.children():
            if isinstance(layer, OperatorList):
                layer.transfer_static()
                continue
            
            for name, param in layer.named_parameters():
                param.requires_grad = False
                self.register_parameter(name, param)

    def randomize_param(self, arg0: float = 0, arg1: float = 2 * np.pi, 
                        method: str = 'uniform') -> None:
        r"""Randomize parameters of the list based on the initializer.  
        Current we only support Uniform and Normal initializer. 

        Args:
            arg0: first argument of the initializer. Defaults to 0.
            arg1: first argument of the initializer. Defaults to 2 pi.
            method: The sampling method. Defaults to 'uniform'.
        """
        assert method in {
            "uniform",
            "normal",
        }, "The initializer should be uniform or normal."

        for layer in self.children():
            if isinstance(layer, OperatorList):
                layer.randomize_param(arg0, arg1, method)
                continue
            
            for name, param in layer.named_parameters():

                if method == "normal":
                    new_param = Parameter(
                        torch.normal(
                            mean=arg0,
                            std=arg1,
                            size=param.shape,
                            dtype=param.dtype,
                            device=param.device,
                        )
                    )
                elif method == "uniform":
                    new_param = Parameter(
                        torch.rand(param.shape, dtype=param.dtype, device=param.device)
                        * (arg1 - arg0) + arg0
                    )
                else:
                    raise NotImplementedError
                
                layer.register_parameter(name, new_param)
    
    
    
    def __call__(self, state: Optional[State] = None) -> State:
        r"""Same as forward of Neural Network
        """
        return self.forward(state)

    def forward(self, state: Optional[State] = None) -> State:
        r"""Passing a physical input state.

        Args:
            state: initial state. Defaults to zero state.

        Returns:
            output quantum state
        """
        if state is None:
            assert self.num_systems == max(self.system_idx) + 1, \
                "Cannot provide default input state: dimensions of some physical systems are not specified."
            state = zero_state(self.num_systems, self.system_dim)
        
        if len(self) == 0:
            warnings.warn(
                "The operator list is empty: return the input state.", UserWarning)
            return state

        assert self.num_systems <= state.num_systems, \
            f"Insufficient system: received {state.num_systems}, expected >= {self.num_systems}"
        assert self.system_dim == (applied_dim := [state.system_dim[i] for i in self.system_idx]), \
            f"Dimension for systems {self.system_idx} does not agree: received {applied_dim}, expected {self.system_dim}"

        state = state.clone()
        state = super().forward(state)
        return state
    
    @property
    def matrix(self) -> torch.Tensor:
        r"""Get the unitary matrix form of the operator list.
        """
        dim = int(np.prod(self.system_dim))
        input_basis = std_basis(self.num_systems, self.system_dim)
        input_basis._switch_unitary_matrix = True
        
        physical_idx, self.system_idx = self.system_idx, list(range(self.num_systems))
        output = self.forward(input_basis)
        self.system_idx = physical_idx
        
        assert output.backend == 'state_vector', \
            f"The circuit seems to be a noisy circuit: expect 'state_vector', output {output_basis.backend}"
        input_basis = input_basis.bra.view([dim, 1, 1, dim])
        output_basis = output.ket.view([dim, -1, dim, 1])
        return torch.sum(output_basis @ input_basis, dim=0).view(output.batch_dim[1:] + [dim, dim])

    def unitary_matrix(self) -> torch.Tensor:
        r"""Get the unitary matrix form of the circuit.

        Returns:
            Unitary matrix form of the circuit.
        
        """
        warnings.warn(
            "Starting from QuAIRKit 0.4.0, it is recommended to use 'Circuit.matrix' instead of "
            "'Circuit.unitary_matrix' to call the unitary matrix of the circuit.", FutureWarning)
        return self.matrix

    @property
    def depth(self) -> int:
        r"""Depth of gate sequences.
        
        Returns:
            depth of this circuit
        
        Note:
            The measurement is omitted, and all gates are assumed to have depth 1. 
            See Niel's answer in the [StackExchange](https://quantumcomputing.stackexchange.com/a/5772).
        
        """
        system_depth = np.array([0] * self.num_systems)
        for gate_info in self.operator_history:
            sys_idx = gate_info['system_idx']
            if isinstance(sys_idx, int):
                system_depth[sys_idx] += 1
            else:
                system_depth[sys_idx] = np.max(system_depth[sys_idx]) + 1
        return int(np.max(system_depth))
    
    @property
    def qasm(self) -> str:
        r"""String representation of the circuit in OpenQASM-like format.
        
        TODO: remove \n
        """
        qasm_str = ''
        for op in self.children():
            if isinstance(op, OperatorList):
                qasm_str += op.qasm
            else:
                qasm_str += '\n' + op.info.qasm
        return qasm_str 

    def __str__(self):
        return self.qasm


class Layer(OperatorList):
    r"""Base class for built-in trainable quantum circuit ansatz.
    
    Args:
        physical_idx: Physical indices of the systems on which this layer is applied.
        depth: Depth of the layer.
        name: Name of the layer. Defaults to 'Layer'.
        system_dim: Dimension of the systems. Defaults to be qubit-systems.
        
    Note:
        A Circuit instance needs to extend this Layer instance to be used in a circuit. 
    
    """
    def __init__(self, physical_idx: List[int],
                 depth: int, name: str = 'Layer',
                 system_dim: Union[int, List[int]] = 2) -> None:
        assert len(physical_idx) > 1, \
            f"Acted systems in a built-in layer needs more than 1: received {len(physical_idx)}."
        
        self.name = name
        self._depth = depth
        super().__init__(len(physical_idx), system_dim, physical_idx)
    
    def get_latex_name(self, style: str = 'standard') -> str:
        r"""Return the LaTeX name of the layer.
        
        Args:
            style: the style of the plot, can be 'standard', 'compact' or 'detailed'. Defaults to ``standard``.
        """
        depth, name = self._depth, self.name
        if depth > 1:
            if style == 'compact':
                name += r'$\times ' + str(depth) + r'$'
            else:
                name = f"{depth} {name}s"
        return name
        
    def _assert_qubits(self) -> None:
        r"""Assert the input systems are qubits.
        """
        assert self.equal_dim and self.system_dim[0] == 2, \
            f"Incorrect dimension for system {self.system_idx}: received {self.system_dim}, expected all qubits."
    
    def _format_layer_param(self, param: Optional[torch.Tensor], param_shape: List[int]) -> torch.Tensor:
        r"""Format the input parameters for the layer.
        
        Args:
            param: Initial parameters for the layer. Defaults to be self-generated.
            param_shape: Shape of the parameters.
        
        Returns:
            formatted parameters
        """
        if param is None:
            param = torch.rand(param_shape) * 2 * np.pi
        else:
            assert param.numel() == np.prod(param_shape), \
                f"Number of parameters does not match: received {param.shape}, expected size {np.prod(param_shape)}."
        return param
    
    def _get_drawer(self, style: str, decimal: int) -> OperatorListDrawer:
        history = self.operator_history
        if len(history) > 15 and style == 'standard':
            warnings.warn(
                "The circuit may be too large to be readable. " +
                "Will transfer to compact style. You may enforce this by setting style as 'detailed'", UserWarning)
            style = 'compact'
        
        drawer = OperatorListDrawer(style, decimal)
        drawer = drawer.draw_layer(history, self.get_latex_name(style), self._depth)
        return drawer
    
    def plot(self, style: str = 'standard', decimal: int = 2,
             dpi: int = 300, print_code: bool = False, show_plot: bool = True) -> None:
        r"""Plot the circuit layer using LaTeX
        
        Args:
            style: the style of the plot, can be 'standard', 'compact' or 'detailed'. Defaults to ``standard``.
            decimal: number of decimal places to display. Defaults to 2.
            dpi: dots per inches of plot image. Defaults to 300.
            print_code: whether print the LaTeX code of the circuit, default to ``True``.
            show_plot: whether show the plotted circuit, default to ``True``.
        """
        physical_idx, self.system_idx = self.system_idx, list(range(self.num_systems))
        
        drawer = self._get_drawer(style, decimal)
        drawer.fill_all()
        drawer.add_end()
        
        self.system_idx = physical_idx
        
        _fig = drawer.plot(dpi, print_code)
        if show_plot:
            _display_png(_fig)
